from datetime import datetime as dt

import pytest

from cg.constants.subject import Sex
from cg.services.illumina.data_transfer.models import IlluminaFlowCellDTO
from cg.store.models import (
    ApplicationVersion,
    Collaboration,
    Customer,
    Organism,
    Sample,
    User,
    IlluminaFlowCell,
)
from cg.store.store import Store


def test_add_collaboration(store: Store):
    # GIVEN an empty database
    collaboration_query = store._get_query(table=Collaboration)
    assert collaboration_query.first() is None
    internal_id, name = "cust_group", "Test customer group"

    # WHEN adding a new customer group
    new_collaboration = store.add_collaboration(internal_id=internal_id, name=name)
    store.session.add(new_collaboration)
    store.session.commit()

    # THEN it should be stored in the database
    assert collaboration_query.first() == new_collaboration


def test_add_user(store: Store):
    # GIVEN a database with a customer in it that we can connect the user to
    customer = store.add_customer(
        internal_id="custtest",
        name="Test Customer",
        scout_access=False,
        invoice_address="dummy street 1",
        invoice_reference="dummy nr",
    )
    store.session.add(customer)

    # WHEN adding a new user
    name, email = "Paul T. Anderson", "paul.anderson@magnolia.com"
    new_user = store.add_user(customer=customer, email=email, name=name)

    store.session.add(new_user)
    store.session.commit()

    # THEN it should be stored in the database
    assert store._get_query(table=User).first() == new_user


def test_add_microbial_sample(base_store: Store, helpers):
    # GIVEN an empty database
    sample_query = base_store._get_query(table=Sample)
    assert sample_query.first() is None
    customer_obj = helpers.ensure_customer(base_store)
    assert customer_obj
    name = "microbial_sample"
    organism_name = "e. coli"
    internal_id = "lims-id"
    reference_genome = "ref_gen"
    priority = "research"
    application_version = base_store._get_query(table=ApplicationVersion).first()
    base_store.add_organism(organism_name, organism_name, reference_genome)
    organism = base_store._get_query(table=Organism).first()

    # WHEN adding a new microbial sample
    new_sample = base_store.add_sample(
        name=name,
        sex=Sex.UNKNOWN,
        internal_id=internal_id,
        priority=priority,
        application_version=application_version,
        organism=organism,
        reference_genome=reference_genome,
    )
    new_sample.customer = customer_obj
    base_store.session.add(new_sample)
    base_store.session.commit()

    # THEN it should be stored in the database
    assert sample_query.first() == new_sample
    stored_microbial_sample = sample_query.first()
    assert stored_microbial_sample.name == name
    assert stored_microbial_sample.internal_id == internal_id
    assert stored_microbial_sample.reference_genome == reference_genome
    assert stored_microbial_sample.application_version == application_version
    assert stored_microbial_sample.priority_human == priority
    assert stored_microbial_sample.organism == organism


def test_add_pool(rml_pool_store: Store):
    """Tests whether new pools are invoiced as default."""
    # GIVEN a valid customer and a valid application_version
    customer: Customer = rml_pool_store.get_customers()[0]
    application = rml_pool_store.get_application_by_tag(tag="RMLP05R800")
    app_version = (
        rml_pool_store._get_query(table=ApplicationVersion)
        .filter(ApplicationVersion.application_id == application.id)
        .first()
    )

    # WHEN adding a new pool
    new_pool = rml_pool_store.add_pool(
        customer=customer,
        name="pool2",
        order="123456",
        ordered=dt.now(),
        application_version=app_version,
    )

    rml_pool_store.session.add(new_pool)
    rml_pool_store.session.commit()
    # THEN the new pool should have no_invoice = False
    pool = rml_pool_store.get_pool_by_entry_id(entry_id=2)
    assert pool.no_invoice is False


def test_add_illumina_flow_cell(
    illumina_flow_cell_dto: IlluminaFlowCellDTO,
    illumina_flow_cell: IlluminaFlowCell,
    illumina_flow_cell_internal_id: str,
    store: Store,
):
    # GIVEN an Illumina flow cell not in store
    assert not store.get_illumina_flow_cell_by_internal_id(illumina_flow_cell_internal_id)

    # WHEN adding the flow cell to the store
    store.add_illumina_flow_cell(illumina_flow_cell_dto)

    # THEN it should be stored in the database
    assert (
        store.get_illumina_flow_cell_by_internal_id(illumina_flow_cell_internal_id).internal_id
        == illumina_flow_cell_dto.internal_id
    )


def test_add_illumina_flow_cell_already_exists(
    illumina_flow_cell_dto: IlluminaFlowCellDTO, illumina_flow_cell_internal_id: str, store: Store
):
    # GIVEN an Illumina flow cell not in store
    store.add_illumina_flow_cell(illumina_flow_cell_dto)
    assert store.get_illumina_flow_cell_by_internal_id(illumina_flow_cell_internal_id)

    # WHEN adding the flow cell to the store
    # THEN a ValueError should be raised
    with pytest.raises(ValueError):
        store.add_illumina_flow_cell(illumina_flow_cell_dto)



"""Class with small helper functions"""

from typing import Iterable


class SmallHelpers:
    """Hold small methods that might be helpful for the tests"""

    @staticmethod
    def length_of_iterable(iter_obj: Iterable) -> int:
        """Returns the length of an iterable"""
        return sum(1 for _ in iter_obj)

    @staticmethod
    def length_of_nested_list(nested_list: list[list]) -> int:
        """Returns the length of a nested list"""
        return sum(1 for _ in nested_list for _ in _)



"""Filters for the Illumina Sequencing Run objects."""

from enum import Enum

from sqlalchemy.orm import Query

from cg.store.models import IlluminaFlowCell, IlluminaSequencingRun


def filter_by_device_internal_id(runs: Query, device_internal_id: str, **kwargs) -> Query:
    """Filter sequencing runs by device internal id."""
    joined_query: Query = runs.join(IlluminaFlowCell)
    return joined_query.filter(IlluminaFlowCell.internal_id == device_internal_id)


def filter_by_entry_id(runs: Query, entry_id: int, **kwargs) -> Query:
    """Filter sequencing runs by entry id."""
    return runs.filter(IlluminaSequencingRun.id == entry_id)


def filter_by_data_availability(runs: Query, data_availability: list[str], **kwargs) -> Query:
    """Filter sequencing runs by data availability statuses."""
    return runs.filter(IlluminaSequencingRun.data_availability.in_(data_availability))


class IlluminaSequencingRunFilter(Enum):
    BY_DEVICE_INTERNAL_ID: callable = filter_by_device_internal_id
    BY_ENTRY_ID: callable = filter_by_entry_id
    WITH_DATA_AVAILABILITY: callable = filter_by_data_availability


def apply_illumina_sequencing_run_filter(
    runs: Query,
    filter_functions: list[callable],
    device_internal_id: str | None = None,
    entry_id: int | None = None,
    data_availability: list[str] | None = None,
) -> Query:
    for filter_function in filter_functions:
        runs: Query = filter_function(
            runs=runs,
            device_internal_id=device_internal_id,
            entry_id=entry_id,
            data_availability=data_availability,
        )
    return runs



"""Test how the api handles versions."""

import datetime
from typing import Any

from cg.apps.housekeeper.hk import HousekeeperAPI
from tests.small_helpers import SmallHelpers


def test_new_version(housekeeper_api: HousekeeperAPI, timestamp: datetime.datetime):
    """Test to create a new version with the API."""
    # GIVEN a housekeeper api and a date

    # WHEN creating a version object
    version_obj = housekeeper_api.new_version(created_at=timestamp)

    # THEN assert that the version obj was created
    assert version_obj

    # THEN assert that the version object has the correct date
    assert version_obj.created_at == timestamp

    # THEN assert that the other members are set to None
    for member in [
        version_obj.expires_at,
        version_obj.included_at,
        version_obj.archived_at,
        version_obj.archive_path,
    ]:
        assert member is None

    # THEN assert that there is no bundle id
    assert version_obj.bundle_id is None


def test_get_version_non_existing(housekeeper_api: HousekeeperAPI, timestamp: datetime.datetime):
    """Test to get a version when there are no existing version."""
    # GIVEN a empty housekeeper_api

    # WHEN fetching a version
    version_obj = housekeeper_api.version(bundle=None, date=timestamp)

    # THEN assert that the function returns None
    assert version_obj is None


def test_get_version_existing(
    housekeeper_api: HousekeeperAPI,
    timestamp_yesterday: datetime.datetime,
    hk_bundle_data: dict[str, Any],
):
    """Test to get a version when there is a bundle and a version."""
    # GIVEN a populated housekeeper_api
    bundle_obj, version_obj = housekeeper_api.add_bundle(hk_bundle_data)
    housekeeper_api.add_commit(bundle_obj)
    housekeeper_api.add_commit(version_obj)

    # WHEN fetching a version
    fetched_version = housekeeper_api.version(bundle=bundle_obj.name, date=timestamp_yesterday)

    # THEN assert that the function returns True
    assert fetched_version

    # THEN assert version has a bundle_id attribute
    assert fetched_version.bundle_id == bundle_obj.id


def test_add_version_existing_bundle(
    populated_housekeeper_api: HousekeeperAPI,
    later_timestamp: datetime.datetime,
    case_id: str,
    small_helpers: SmallHelpers,
):
    """Test to get a version when there is a bundle and a version."""
    # GIVEN a populated housekeeper_api and a bundle with one version
    bundle_obj = populated_housekeeper_api.bundle(name=case_id)
    assert bundle_obj
    assert small_helpers.length_of_iterable(bundle_obj.versions) == 1

    # WHEN creating a newer version and adding it to the bundle
    new_version = populated_housekeeper_api.new_version(created_at=later_timestamp)
    new_version.bundle = bundle_obj
    populated_housekeeper_api.add_commit(new_version)

    # THEN assert that the bundle has two versions
    bundle_obj = populated_housekeeper_api.bundle(name=case_id)
    assert small_helpers.length_of_iterable(bundle_obj.versions) == 2


def test_get_last_version(
    case_id: str, populated_housekeeper_api: HousekeeperAPI, timestamp_now: datetime.datetime
):
    """Test to get a version when there is a bundle and a version."""
    # GIVEN a populated housekeeper_api and a bundle with two versions
    bundle_obj = populated_housekeeper_api.bundle(name=case_id)
    new_version = populated_housekeeper_api.new_version(created_at=timestamp_now)
    new_version.bundle = bundle_obj
    populated_housekeeper_api.add_commit(new_version)

    # WHEN fetching the last version of a bundle
    fetched_version = populated_housekeeper_api.last_version(bundle=case_id)

    # THEN assert that the later_date version is fetched
    assert fetched_version.created_at == timestamp_now


def test_get_latest_bundle_version_no_housekeeper_bundle(
    case_id: str, housekeeper_api: HousekeeperAPI, caplog
):
    """Test get_latest_bundle_version function when there is no case bundle in Housekeeper."""

    # GIVEN a case id

    # WHEN getting the version_obj
    res = housekeeper_api.get_latest_bundle_version(bundle_name=case_id)

    # THEN assert that None was returned since there is no such case bundle in Housekeeper
    assert res is None

    # THEN assert that no bundle is found
    assert f"No bundle found for {case_id} in Housekeeper" in caplog.text


def test_get_create_version(
    another_case_id: str,
    populated_housekeeper_api: HousekeeperAPI,
    case_id: str,
):
    # Given a populated housekeeper_api with a bundle
    version_obj = populated_housekeeper_api.bundle(case_id).versions[0]

    # When the case_id is given the function should return its latest version
    latest_version = populated_housekeeper_api.get_or_create_version(case_id)

    # Then assert that the versions match
    assert version_obj == latest_version

    # When a case_id not present in hk is given.
    latest_version = populated_housekeeper_api.get_or_create_version(another_case_id)

    # Then assert the new bundle is created the version is new.
    assert latest_version.bundle.name == another_case_id



from pathlib import Path

import pytest

from cg.constants import FileExtensions
from cg.exc import ValidationError
from cg.io.validate_path import validate_file_suffix


def test_validate_file_suffix_correct_suffix(csv_file_path: Path):
    """Test validate file suffix when file suffix is in correct format."""
    # GIVEN a existing file in the correct file format
    assert csv_file_path.exists()

    # WHEN validating the file suffix
    was_validated: bool = validate_file_suffix(
        path_to_validate=csv_file_path, target_suffix=FileExtensions.CSV
    )

    # THEN assert the suffix is in the correct format
    assert was_validated


def test_validate_file_suffix_wrong_suffix(caplog, json_file_path: Path):
    """Test validate file suffix when file suffix is in wrong format."""
    # GIVEN a existing file in the wrong file format
    assert json_file_path.exists()
    assert json_file_path.suffix != FileExtensions.CSV

    # WHEN validating the file suffix
    with pytest.raises(ValidationError):
        validate_file_suffix(path_to_validate=json_file_path, target_suffix=FileExtensions.CSV)

        # THEN assert the suffix is in the wrong format
        assert "seems to be in wrong format" in caplog.text



""" Start of CLI """

import logging
import sys
from pathlib import Path

import click
import coloredlogs
from sqlalchemy.orm import scoped_session

import cg
from cg.cli.sequencing_qc.sequencing_qc import sequencing_qc
from cg.cli.validate import validate
from cg.cli.add import add as add_cmd
from cg.cli.archive import archive
from cg.cli.backup import backup
from cg.cli.clean import clean
from cg.cli.compress.base import compress, decompress
from cg.cli.delete.base import delete
from cg.cli.deliver.base import deliver as deliver_cmd
from cg.cli.demultiplex.base import demultiplex_cmd_group as demultiplex_cmd
from cg.cli.downsample import downsample
from cg.cli.generate.base import generate as generate_cmd
from cg.cli.get import get
from cg.cli.set.base import set_cmd
from cg.cli.store.base import store as store_cmd
from cg.cli.transfer import transfer_group
from cg.cli.upload.base import upload
from cg.cli.utils import CLICK_CONTEXT_SETTINGS
from cg.cli.validate import validate
from cg.cli.workflow.base import workflow as workflow_cmd
from cg.constants.cli_options import FORCE
from cg.constants.constants import FileFormat
from cg.io.controller import ReadFile
from cg.models.cg_config import CGConfig
from cg.store.database import (
    create_all_tables,
    drop_all_tables,
    get_scoped_session_registry,
    get_tables,
)

LOG = logging.getLogger(__name__)
LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR"]


def teardown_session():
    """Ensure that the session is closed and all resources are released to the connection pool."""
    registry: scoped_session | None = get_scoped_session_registry()
    if registry:
        registry.remove()


@click.group(context_settings=CLICK_CONTEXT_SETTINGS)
@click.option("-c", "--config", type=click.Path(exists=True), help="path to config file")
@click.option("-d", "--database", help="path/URI of the SQL database")
@click.option(
    "-l", "--log-level", type=click.Choice(LEVELS), default="INFO", help="lowest level to log at"
)
@click.option("--verbose", is_flag=True, help="Show full log information, time stamp etc")
@click.version_option(cg.__version__, prog_name=cg.__title__)
@click.pass_context
def base(
    context: click.Context,
    config: click.Path,
    database: str | None,
    log_level: str,
    verbose: bool,
):
    """cg - interface between tools at Clinical Genomics."""
    if verbose:
        log_format = "%(asctime)s %(hostname)s %(name)s[%(process)d] %(levelname)s %(message)s"
    else:
        log_format = "%(message)s" if sys.stdout.isatty() else None

    coloredlogs.install(level=log_level, fmt=log_format)
    raw_config: dict = (
        ReadFile.get_content_from_file(file_format=FileFormat.YAML, file_path=Path(config))
        if config
        else {"database": database}
    )
    context.obj = CGConfig(**raw_config)
    context.call_on_close(teardown_session)


@base.command()
@click.option("--reset", is_flag=True, help="Reset database before setting up tables")
@FORCE
@click.pass_obj
def init(context: CGConfig, reset: bool, force: bool):
    """Setup the database."""
    existing_tables: list[str] = get_tables()
    if force or reset:
        if existing_tables and not force:
            message = f"Delete existing tables? [{', '.join(existing_tables)}]"
            click.confirm(click.style(message, fg="yellow"), abort=True)
        drop_all_tables()
    elif existing_tables:
        LOG.error("Database already exists, use '--reset'")
        raise click.Abort

    create_all_tables()
    LOG.info(f"Success! New tables: {', '.join(get_tables())}")


base.add_command(add_cmd)
base.add_command(archive)
base.add_command(backup)
base.add_command(clean)
base.add_command(compress)
base.add_command(decompress)
base.add_command(delete)
base.add_command(get)
base.add_command(set_cmd)
base.add_command(transfer_group)
base.add_command(upload)
base.add_command(workflow_cmd)
base.add_command(store_cmd)
base.add_command(deliver_cmd)
base.add_command(demultiplex_cmd)
base.add_command(generate_cmd)
base.add_command(downsample)
base.add_command(validate)
base.add_command(sequencing_qc)



from cg.constants.nipt import Q30_THRESHOLD
from cg.meta.upload.nipt import NiptUploadAPI
from cg.models.cg_config import CGConfig


def test_sequencing_run_passed_qc_value(
    case_id_for_sample_on_multiple_flow_cells: str, nipt_upload_api_context: CGConfig
):
    # GIVEN a NiptUploadAPI context with a case that has been sequenced
    nipt_upload_api: NiptUploadAPI = NiptUploadAPI(nipt_upload_api_context)

    # WHEN checking the Q30 and read values on the flow cell
    sequencing_run_qc_value: bool = nipt_upload_api.sequencing_run_passed_qc_value(
        case_id=case_id_for_sample_on_multiple_flow_cells, q30_threshold=Q30_THRESHOLD
    )

    # THEN the successful flow cell shall pass
    assert sequencing_run_qc_value


def test_sequencing_run_failed_qc_value(
    nipt_upload_api_failed_fc_context: CGConfig,
    case_id_for_sample_on_multiple_flow_cells: str,
    sample_id: str,
):
    # GIVEN a NiptUploadAPI context with a case that has been sequenced, however the case is lacking reads
    nipt_upload_api: NiptUploadAPI = NiptUploadAPI(nipt_upload_api_failed_fc_context)

    # WHEN checking the Q30 and read values on the sequencing run
    sequencing_run_qc_pass: bool = nipt_upload_api.sequencing_run_passed_qc_value(
        case_id=case_id_for_sample_on_multiple_flow_cells, q30_threshold=Q30_THRESHOLD
    )

    # THEN the failed sequencing run should have not been passed
    assert not sequencing_run_qc_pass



"""Test Tomte delivery report API."""

from cg.meta.report.tomte import TomteReportAPI
from cg.models.analysis import NextflowAnalysis
from cg.models.report.metadata import TomteSampleMetadataModel
from cg.store.models import Case, Sample


def test_get_sample_metadata(
    report_api_tomte: TomteReportAPI,
    sample_id: str,
    tomte_case_id: str,
    tomte_validated_metrics: dict[str, str],
    tomte_mock_analysis_finish,
):
    """Test Tomte sample metadata extraction."""

    # GIVEN a Tomte case and associated sample
    case: Case = report_api_tomte.status_db.get_case_by_internal_id(internal_id=tomte_case_id)
    sample: Sample = report_api_tomte.status_db.get_sample_by_internal_id(internal_id=sample_id)

    # GIVEN an analysis metadata object
    latest_metadata: NextflowAnalysis = report_api_tomte.analysis_api.get_latest_metadata(
        case_id=tomte_case_id
    )

    # WHEN getting the sample metadata
    sample_metadata: TomteSampleMetadataModel = report_api_tomte.get_sample_metadata(
        case=case, sample=sample, analysis_metadata=latest_metadata
    )

    # THEN the sample metadata should be correctly retrieved and match the expected metrics
    assert sample_metadata.model_dump() == tomte_validated_metrics



import re

import click

import shutil


def echo_lines(lines: list[str]) -> None:
    for line in lines:
        click.echo(line)


def is_case_name_allowed(name: str) -> bool:
    """Returns true if the given name consists only of letters, numbers and dashes."""
    allowed_pattern: re.Pattern = re.compile("^[A-Za-z0-9-]+$")
    return bool(allowed_pattern.fullmatch(name))


CLICK_CONTEXT_SETTINGS: dict[str, int] = {
    "max_content_width": shutil.get_terminal_size().columns - 10
}



"""Tests for the illumina clean runs service."""

import logging
import time

import mock
import pytest
from _pytest.logging import LogCaptureFixture
from housekeeper.store.models import File

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import SequencingFileTag
from cg.constants.time import TWENTY_ONE_DAYS_IN_SECONDS
from cg.exc import IlluminaCleanRunError, HousekeeperFileMissingError
from cg.services.illumina.cleaning.clean_runs_service import (
    IlluminaCleanRunsService,
)
from cg.store.models import (
    Sample,
    IlluminaSequencingRun,
    IlluminaSampleSequencingMetrics,
)
from tests.store_helpers import StoreHelpers


def test_get_sequencing_run_from_statusdb(
    illumina_clean_service_can_be_removed: IlluminaCleanRunsService,
):
    """Test to get a sequencing run from statusdb."""

    # GIVEN an Illumina clean sequencing runs service with a store that contains a sequencing run

    # WHEN retrieving the sequencing run from the store
    sequencing_run: IlluminaSequencingRun = (
        illumina_clean_service_can_be_removed.get_sequencing_run_from_status_db()
    )

    # THEN a sequencing run is retrieved
    assert isinstance(sequencing_run, IlluminaSequencingRun)


def test_is_sequencing_run_in_statusdb(
    illumina_clean_service_can_be_removed: IlluminaCleanRunsService,
):
    """Test to check that a sequencing run is in statusdb."""
    # GIVEN a clean illumina clean sequencing runs service with a store that contains a sequencing run

    # WHEN checking whether a sequencing run is in status DB
    is_in_statusdb: bool = illumina_clean_service_can_be_removed.is_sequencing_run_in_statusdb()

    # THEN checking whether a sequencing run is in statusdb is TRUE
    assert is_in_statusdb


def test_is_sequencing_run_backed_up(
    illumina_clean_service_can_be_removed: IlluminaCleanRunsService,
):
    """Test to check that a sequencing run has been backed up."""
    # GIVEN a clean illumina clean sequencing runs service with a sequencing run that has been backed up

    # WHEN checking whether the sequencing run has been backed
    is_backed_up: bool = illumina_clean_service_can_be_removed.is_sequencing_run_backed_up()

    # THEN checking whether the sequencing run has been backed up is TRUE
    assert is_backed_up


def test_is_not_sequencing_run_backed_up(
    illumina_clean_service_can_be_removed: IlluminaCleanRunsService,
):
    """Test to check that a sequencing run has been backed up."""
    # GIVEN a clean illumina clean sequencing runs service with a sequencing run that has not been backed up
    illumina_clean_service_can_be_removed.get_sequencing_run_from_status_db().has_backup = False

    # WHEN checking whether the sequencing run has been backed
    is_backed_up: bool = illumina_clean_service_can_be_removed.is_sequencing_run_backed_up()

    # THEN checking whether the sequencing run has been backed up is FALSE
    assert not is_backed_up


def test_get_sequencing_metrics_for_sequencing_run_from_statusdb(
    illumina_clean_service_can_be_removed: IlluminaCleanRunsService,
):
    """Test to get a sequencing metrics for a sequencing run from statusdb."""

    # GIVEN a illumina clean sequencing runs service with a store that contains a sequencing run with IlluminaSampleSequencingMetrics entries

    # WHEN retrieving the sequencing run from the store
    metrics: list[IlluminaSampleSequencingMetrics] = (
        illumina_clean_service_can_be_removed.get_sequencing_metrics_for_sequencing_run()
    )

    # THEN a IlluminaSampleSequencingMetrics entry is retrieved
    assert isinstance(metrics[0], IlluminaSampleSequencingMetrics)


def test_has_sequencing_metrics_in_statusdb(
    illumina_clean_service_can_be_removed: IlluminaCleanRunsService,
):
    """Test to check that a sequencing run has IlluminaSampleSequencingMetrics entries in statusdb."""

    # GIVEN an Illumina clean sequencing runs service with a store that contains a sequencing run with IlluminaSampleSequencingMetrics entries

    # WHEN THEN checking whether a sequencing run has IlluminaSampleSequencingMetrics entries'
    has_metrics: bool = illumina_clean_service_can_be_removed.has_sequencing_metrics_in_statusdb()

    # THEN checking whether a sequencing run has IlluminaSampleSequencingMetrics entries returns True
    assert has_metrics


def test_is_directory_older_than_21_days_pass(
    illumina_clean_service_can_be_removed: IlluminaCleanRunsService,
):
    """Test to check whether a directory is older than 21 days."""

    # GIVEN an Illumina clean sequencing runs service with a sequencing run that can be deleted

    # WHEN checking whether a given sequencing run directory is older than 21 days

    with mock.patch(
        "time.time",
        return_value=time.time() + TWENTY_ONE_DAYS_IN_SECONDS,
    ):
        is_older_that_21_days: bool = (
            illumina_clean_service_can_be_removed.is_directory_older_than_21_days()
        )

    # THEN checking whether a given sequencing run directory is older than 21 days is TRUE
    assert is_older_that_21_days


def test_is_directory_older_than_21_days_fail(
    illumina_clean_service_can_be_removed: IlluminaCleanRunsService,
):
    """Test to check whether a directory is older than 21 days."""

    # GIVEN an Illumina clean sequencing runs service with a current time that is set to now.

    # WHEN checking whether a given sequencing run directory is older than 21 days
    is_older_than_21_days: bool = (
        illumina_clean_service_can_be_removed.is_directory_older_than_21_days()
    )

    # THEN checking whether a given sequencing run directory is older than 21 days is FALSE
    assert not is_older_than_21_days


def test_has_sample_sheet_in_housekeeper(
    illumina_clean_service_can_be_removed: IlluminaCleanRunsService,
):
    """Test to check whether a sequencing run has a sample sheet in housekeeper."""

    # GIVEN an Illumina clean sequencing runs service with a sequencing run that has a sample sheet in housekeeper

    # WHEN checking whether the sequencing run has a sample sheet in housekeeper
    has_sample_sheet: bool = illumina_clean_service_can_be_removed.has_sample_sheet_in_housekeeper()

    # THEN checking whether the sequencing run has a sample sheet in housekeeper returns TRUE
    assert has_sample_sheet


@pytest.mark.parametrize(
    "tag",
    [SequencingFileTag.FASTQ, SequencingFileTag.SPRING, SequencingFileTag.SPRING_METADATA],
    ids=["fastq", "spring", "spring_metadata"],
)
def test_get_files_for_flow_cell_bundle(
    illumina_clean_service_can_be_removed: IlluminaCleanRunsService, tag: str
):
    """Test to get files for a sequencing run bundle from housekeeper."""

    # GIVEN an Illumina clean sequencing runs service with a sequencing run that has files in housekeeper

    # WHEN getting the sequencing run samples' files
    files: list[File] = (
        illumina_clean_service_can_be_removed.get_files_for_samples_on_flow_cell_with_tag(tag=tag)
    )

    # THEN files are returned
    assert files


@pytest.mark.parametrize(
    "tag",
    [SequencingFileTag.FASTQ, SequencingFileTag.SPRING, SequencingFileTag.SPRING_METADATA],
    ids=["fastq", "spring", "spring_metadata"],
)
def test_get_files_for_samples_on_flow_cell_with_tag_missing_sample(
    illumina_clean_service_can_not_be_removed: IlluminaCleanRunsService,
    helpers: StoreHelpers,
    tag: str,
    selected_novaseq_6000_pre_1_5_kits_sample_ids: list[str],
    caplog: LogCaptureFixture,
):
    """Test that a sequencing run is cleaned if one of its samples is missing in Housekeeper."""
    caplog.set_level(logging.WARNING)

    # GIVEN a store with a sequencing run to be cleaned

    # GIVEN that the sequencing run has two samples
    first_sample: Sample = (
        illumina_clean_service_can_not_be_removed.status_db.get_sample_by_internal_id(
            selected_novaseq_6000_pre_1_5_kits_sample_ids[0]
        )
    )
    second_sample: Sample = (
        illumina_clean_service_can_not_be_removed.status_db.get_sample_by_internal_id(
            selected_novaseq_6000_pre_1_5_kits_sample_ids[1]
        )
    )
    assert first_sample
    assert second_sample

    # GIVEN that one of the samples is not in Housekeeper
    hk_api: HousekeeperAPI = illumina_clean_service_can_not_be_removed.hk_api
    assert not hk_api.bundle(second_sample.internal_id)

    # WHEN getting the sequencing run samples' files
    files: list[File] = (
        illumina_clean_service_can_not_be_removed.get_files_for_samples_on_flow_cell_with_tag(
            tag=tag
        )
    )

    # THEN files are returned
    assert not files

    # THEN a warning is logged
    assert f"Bundle: {first_sample.internal_id} not found in Housekeeper" in caplog.text


def test_can_sequencing_run_be_deleted(
    illumina_clean_service_can_be_removed: IlluminaCleanRunsService,
):
    """Test the sequencing run can be deleted check."""
    # GIVEN a sequencing run that can be deleted

    with mock.patch(
        "cg.services.illumina.cleaning.clean_runs_service.IlluminaCleanRunsService.is_directory_older_than_21_days",
        return_value=True,
    ):
        # WHEN checking that the sequencing run can be deleted
        can_be_deleted: bool = illumina_clean_service_can_be_removed.can_run_directory_be_deleted()

    # THEN the check whether the sequencing run can be deleted returns True
    assert can_be_deleted


def test_can_sequencing_run_be_deleted_no_spring_with_fastq(
    illumina_clean_service_can_be_removed: IlluminaCleanRunsService,
):
    """Test that a sequencing run can be deleted when it has fastq files but no spring files."""
    # GIVEN a sequencing run that can be deleted

    with mock.patch(
        "cg.services.illumina.cleaning.clean_runs_service.IlluminaCleanRunsService.is_directory_older_than_21_days",
        return_value=True,
    ):
        with mock.patch(
            "cg.services.illumina.cleaning.clean_runs_service.IlluminaCleanRunsService.has_spring_meta_data_files_for_samples_in_housekeeper",
            return_value=False,
        ):
            # WHEN checking that the sequencing run can be deleted
            can_be_deleted: bool = (
                illumina_clean_service_can_be_removed.can_run_directory_be_deleted()
            )

    # THEN the check whether the sequencing run can be deleted returns True
    assert can_be_deleted


def test_can_sequencing_run_be_deleted_spring_no_fastq(
    illumina_clean_service_can_be_removed: IlluminaCleanRunsService,
):
    """Test that a sequencing run can be deleted when it has spring files but no fastq files."""
    # GIVEN a sequencing run that can be deleted

    with mock.patch(
        "cg.services.illumina.cleaning.clean_runs_service.IlluminaCleanRunsService.is_directory_older_than_21_days",
        return_value=True,
    ):
        with mock.patch(
            "cg.services.illumina.cleaning.clean_runs_service.IlluminaCleanRunsService.has_fastq_files_for_samples_in_housekeeper",
            return_value=False,
        ):
            # WHEN checking that the sequencing run can be deleted
            can_be_deleted: bool = (
                illumina_clean_service_can_be_removed.can_run_directory_be_deleted()
            )

    # THEN the check whether the sequencing run can be deleted returns True
    assert can_be_deleted


def test_can_sequencing_run_be_deleted_no_spring_no_fastq(
    illumina_clean_service_can_be_removed: IlluminaCleanRunsService,
):
    """Test that a sequencing run can not be deleted when it has no spring files and no fastq files."""
    # GIVEN a sequencing run that can be deleted
    with mock.patch(
        "cg.services.illumina.cleaning.clean_runs_service.IlluminaCleanRunsService.is_directory_older_than_21_days",
        return_value=True,
    ), mock.patch(
        "cg.services.illumina.cleaning.clean_runs_service.IlluminaCleanRunsService.has_fastq_files_for_samples_in_housekeeper",
        return_value=False,
    ), mock.patch(
        "cg.services.illumina.cleaning.clean_runs_service.IlluminaCleanRunsService.has_spring_meta_data_files_for_samples_in_housekeeper",
        return_value=False,
    ):
        # WHEN checking that the sequencing run can be deleted

        # THEN a HousekeeperFileMissingError is raised
        with pytest.raises(HousekeeperFileMissingError):
            illumina_clean_service_can_be_removed.can_run_directory_be_deleted()


def test_delete_sequencing_run_directory(
    illumina_clean_service_can_be_removed: IlluminaCleanRunsService,
):
    """Test that a sequencing run directory is removed."""
    # GIVEN a sequencing run that can be removed

    # GIVEN that the sequencing run directory exists
    assert illumina_clean_service_can_be_removed.sequencing_run_dir_data.path.exists()

    # WHEN removing the sequencing run directory
    with mock.patch(
        "cg.services.illumina.cleaning.clean_runs_service.IlluminaCleanRunsService.is_directory_older_than_21_days",
        return_value=True,
    ):
        illumina_clean_service_can_be_removed.delete_run_directory()

    # THEN the sequencing run directory is removed
    assert not illumina_clean_service_can_be_removed.sequencing_run_dir_data.path.exists()


def test_delete_sequencing_run_directory_can_not_be_deleted(
    illumina_clean_service_can_not_be_removed: IlluminaCleanRunsService,
):
    """Test delete a sequencing run that does not pass all checks."""
    # GIVEN a sequencing run that should not be removed.

    # GIVEN that the sequencing run directory exists
    assert illumina_clean_service_can_not_be_removed.sequencing_run_dir_data.path.exists()

    # WHEN trying to remove the sequencing run
    with pytest.raises(IlluminaCleanRunError):
        illumina_clean_service_can_not_be_removed.delete_run_directory()

    # THEN the sequencing run directory still exists
    illumina_clean_service_can_not_be_removed.sequencing_run_dir_data.path.exists()


def test_get_sequencing_run_from_statusdb_does_not_exist(
    illumina_clean_service_can_not_be_removed: IlluminaCleanRunsService,
):
    """Test retrieving a sequencing run from statusDB that does not exist."""
    # GIVEN a illumina clean sequencing runs service with a sequencing run that is not in statusDB
    illumina_clean_service_can_not_be_removed.sequencing_run_dir_data.id = (
        "flow_cell_does_not_exist"
    )

    assert not illumina_clean_service_can_not_be_removed.status_db.get_illumina_sequencing_run_by_device_internal_id(
        illumina_clean_service_can_not_be_removed.sequencing_run_dir_data.id
    )

    # WHEN retrieving the sequencing run from statusDB

    # THEN a ValueError is raised
    with pytest.raises(ValueError):
        illumina_clean_service_can_not_be_removed.get_sequencing_run_from_status_db()


def test_sequencing_run_has_not_fastq_files_in_housekeeper(
    illumina_clean_service_can_not_be_removed: IlluminaCleanRunsService,
    caplog,
):
    """Test to check wether a sequencing run has files in Housekeeper."""
    caplog.set_level(logging.DEBUG)
    # GIVEN a sequencing run that has entries in StatusDB but does not have any files in Housekeeper for its samples

    # WHEN checking whether a sample on a sequencing run has fastq files

    # THEN the sequencing run has no fastq files in housekeeper.
    assert (
        not illumina_clean_service_can_not_be_removed.has_fastq_files_for_samples_in_housekeeper()
    )
    assert SequencingFileTag.FASTQ in caplog.text


def test_sequencing_run_has_not_spring_files_in_housekeeper(
    illumina_clean_service_can_not_be_removed: IlluminaCleanRunsService,
    caplog,
):
    """Test to check whether a sequencing run has files in Housekeeper."""
    caplog.set_level(logging.DEBUG)
    # GIVEN a sequencing run that has entries in StatusDB but does not have any files in Housekeeper for its samples

    # WHEN checking whether a sample on a sequencing run has spring files

    # THEN the sequencing run has no spring files for samples in housekeeper
    assert (
        not illumina_clean_service_can_not_be_removed.has_spring_files_for_samples_in_housekeeper()
    )
    assert SequencingFileTag.SPRING in caplog.text


def test_sequencing_run_has_not_spring_meta_data_files_in_housekeeper(
    illumina_clean_service_can_not_be_removed: IlluminaCleanRunsService,
    caplog,
):
    """Test to check whether a sequencing run has files in Housekeeper."""
    caplog.set_level(logging.DEBUG)
    # GIVEN a sequencing run that has entries in StatusDB but does not have any files in Housekeeper for its samples

    # WHEN checking whether a sample on a sequencing run has spring files

    # THEN the sequencing run has no spring metadata files in housekeeper

    assert (
        not illumina_clean_service_can_not_be_removed.has_spring_meta_data_files_for_samples_in_housekeeper()
    )
    assert SequencingFileTag.SPRING_METADATA in caplog.text


def test_has_no_sample_files_in_housekeeper(
    illumina_clean_service_can_not_be_removed: IlluminaCleanRunsService,
):
    # GIVEN a sequencing run that has entries in StatusDB but does not have any files in Housekeeper for its samples

    # WHEN checking whether a sample on a sequencing run has files in housekeeper

    # THEN a HousekeeperFileMissingError is raised
    with pytest.raises(HousekeeperFileMissingError):
        illumina_clean_service_can_not_be_removed.has_sample_fastq_or_spring_files_in_housekeeper()



"""Handler to update data objects"""

from datetime import datetime

from sqlalchemy.orm import Session


from cg.constants import SequencingRunDataAvailability
from cg.store.base import BaseHandler
from cg.store.models import IlluminaSequencingRun, IlluminaSampleSequencingMetrics, SampleRunMetrics

from cg.constants.constants import SequencingQCStatus
from cg.store.base import BaseHandler
from cg.store.models import Case, Order, Sample


class UpdateHandler(BaseHandler):
    """Contains methods to update database objects."""

    def __init__(self, session: Session):
        super().__init__(session=session)
        self.session = session

    def update_sample_comment(self, sample: Sample, comment: str) -> None:
        """Update comment on sample with the provided comment."""
        if sample.comment:
            sample.comment = sample.comment + " " + comment
        else:
            sample.comment = comment
        self.session.commit()

    def update_order_delivery(self, order_id: int, delivered: bool) -> Order:
        """Update the delivery status of an order."""
        order: Order = self.get_order_by_id(order_id)
        order.is_delivered = delivered
        self.session.commit()
        return order

    def update_illumina_sequencing_run_data_availability(
        self,
        sequencing_run: IlluminaSequencingRun,
        data_availability: SequencingRunDataAvailability,
    ) -> IlluminaSequencingRun:
        """Update the data availability status of an Illumina sequencing run."""
        sequencing_run.data_availability = data_availability
        self.session.commit()
        return sequencing_run

    def update_illumina_sequencing_run_has_backup(
        self, sequencing_run: IlluminaSequencingRun, has_backup: bool
    ) -> IlluminaSequencingRun:
        """Update the backup status of an Illumina sequencing run."""
        sequencing_run.has_backup = has_backup
        self.session.commit()
        return sequencing_run

    def update_sequencing_qc_status(self, case: Case, status: SequencingQCStatus) -> None:
        case.aggregated_sequencing_qc = status
        self.session.commit()

    def update_sample_reads_illumina(self, internal_id: str):
        sample: Sample = self.get_sample_by_internal_id(internal_id)
        total_reads_for_sample: int = 0
        sample_metrics: list[IlluminaSampleSequencingMetrics] = sample.sample_run_metrics
        for sample_metric in sample_metrics:
            total_reads_for_sample += sample_metric.total_reads_in_lane
        sample.reads = total_reads_for_sample
        self.session.commit()

    def update_sample_sequenced_at(self, internal_id: str, date: datetime):
        sample: Sample = self.get_sample_by_internal_id(internal_id)
        sample.last_sequenced_at = date
        self.session.commit()



import pytest

from cg.store.models import Case, Sample
from cg.constants.priority import Priority
from cg.constants.constants import PrepCategory, Workflow

from tests.fixture_plugins.quality_controller_fixtures.sequencing_qc_check_scenario import (
    SequencingQCCheckScenarios,
)


@pytest.fixture(scope="function")
def sequencing_qc_check_scenarios(store, helpers) -> SequencingQCCheckScenarios:
    return SequencingQCCheckScenarios(store=store, helpers=helpers)


@pytest.fixture(scope="function")
def ready_made_library_sample_passing_sequencing_qc(
    sequencing_qc_check_scenarios: SequencingQCCheckScenarios,
) -> Sample:
    return sequencing_qc_check_scenarios.get_sample_scenario(
        priority=Priority.standard,
        prep_category=PrepCategory.READY_MADE_LIBRARY,
        pass_sequencing_qc=True,
    )


@pytest.fixture(scope="function")
def ready_made_library_sample_failing_sequencing_qc(
    sequencing_qc_check_scenarios: SequencingQCCheckScenarios,
) -> Sample:
    return sequencing_qc_check_scenarios.get_sample_scenario(
        priority=Priority.standard,
        prep_category=PrepCategory.READY_MADE_LIBRARY,
        pass_sequencing_qc=False,
    )


@pytest.fixture(scope="function")
def express_sample_passing_sequencing_qc(
    sequencing_qc_check_scenarios: SequencingQCCheckScenarios,
) -> Sample:
    return sequencing_qc_check_scenarios.get_sample_scenario(
        priority=Priority.express,
        prep_category=PrepCategory.WHOLE_GENOME_SEQUENCING,
        pass_sequencing_qc=True,
    )


@pytest.fixture(scope="function")
def express_sample_failing_sequencing_qc(
    sequencing_qc_check_scenarios: SequencingQCCheckScenarios,
) -> Sample:
    return sequencing_qc_check_scenarios.get_sample_scenario(
        priority=Priority.express,
        prep_category=PrepCategory.WHOLE_GENOME_SEQUENCING,
        pass_sequencing_qc=False,
    )


@pytest.fixture(scope="function")
def sample_passing_sequencing_qc(
    sequencing_qc_check_scenarios: SequencingQCCheckScenarios,
) -> Sample:
    return sequencing_qc_check_scenarios.get_sample_scenario(
        priority=Priority.standard,
        prep_category=PrepCategory.WHOLE_GENOME_SEQUENCING,
        pass_sequencing_qc=True,
    )


@pytest.fixture(scope="function")
def sample_failing_sequencing_qc(
    sequencing_qc_check_scenarios: SequencingQCCheckScenarios,
) -> Sample:
    return sequencing_qc_check_scenarios.get_sample_scenario(
        priority=Priority.standard,
        prep_category=PrepCategory.WHOLE_GENOME_SEQUENCING,
        pass_sequencing_qc=False,
    )


@pytest.fixture(scope="function")
def case_passing_sequencing_qc(
    sequencing_qc_check_scenarios: SequencingQCCheckScenarios,
) -> Case:
    return sequencing_qc_check_scenarios.get_case_scenario(
        priority=Priority.standard,
        pass_sequencing_qc=True,
        prep_category=PrepCategory.WHOLE_GENOME_SEQUENCING,
        workflow=Workflow.MIP_DNA,
    )


@pytest.fixture(scope="function")
def case_failing_sequencing_qc(
    sequencing_qc_check_scenarios: SequencingQCCheckScenarios,
) -> Case:
    return sequencing_qc_check_scenarios.get_case_scenario(
        priority=Priority.standard,
        pass_sequencing_qc=False,
        prep_category=PrepCategory.WHOLE_GENOME_SEQUENCING,
        workflow=Workflow.MIP_DNA,
    )


@pytest.fixture(scope="function")
def express_case_passing_sequencing_qc(
    sequencing_qc_check_scenarios: SequencingQCCheckScenarios,
) -> Case:
    return sequencing_qc_check_scenarios.get_case_scenario(
        priority=Priority.express,
        pass_sequencing_qc=True,
        prep_category=PrepCategory.WHOLE_GENOME_SEQUENCING,
        workflow=Workflow.MIP_DNA,
    )


@pytest.fixture(scope="function")
def express_case_failing_sequencing_qc(
    sequencing_qc_check_scenarios: SequencingQCCheckScenarios,
) -> Case:
    return sequencing_qc_check_scenarios.get_case_scenario(
        priority=Priority.express,
        pass_sequencing_qc=False,
        prep_category=PrepCategory.WHOLE_GENOME_SEQUENCING,
        workflow=Workflow.MIP_DNA,
    )


@pytest.fixture(scope="function")
def one_sample_in_case_has_reads(
    sequencing_qc_check_scenarios: SequencingQCCheckScenarios,
) -> Case:
    return sequencing_qc_check_scenarios.get_case_scenario(
        priority=Priority.standard,
        pass_sequencing_qc=True,
        prep_category=PrepCategory.MICROBIAL,
        workflow=Workflow.MICROSALT,
    )


@pytest.fixture(scope="function")
def no_sample_in_case_has_reads(
    sequencing_qc_check_scenarios: SequencingQCCheckScenarios,
) -> Case:
    return sequencing_qc_check_scenarios.get_case_scenario(
        priority=Priority.standard,
        pass_sequencing_qc=False,
        prep_category=PrepCategory.MICROBIAL,
        workflow=Workflow.MICROSALT,
    )



"""Code for uploading observations data via CLI."""

import logging
from datetime import datetime

import click
from sqlalchemy.orm import Query

from cg.cli.upload.observations.utils import get_observations_api
from cg.cli.workflow.commands import ARGUMENT_CASE_ID, OPTION_LOQUSDB_SUPPORTED_WORKFLOW
from cg.constants import EXIT_FAIL, EXIT_SUCCESS
from cg.constants.cli_options import DRY_RUN
from cg.constants.constants import Workflow
from cg.exc import CgError
from cg.meta.observations.balsamic_observations_api import BalsamicObservationsAPI
from cg.meta.observations.mip_dna_observations_api import MipDNAObservationsAPI
from cg.models.cg_config import CGConfig
from cg.store.store import Store

LOG = logging.getLogger(__name__)


@click.command("observations")
@ARGUMENT_CASE_ID
@DRY_RUN
@click.pass_obj
def upload_observations_to_loqusdb(context: CGConfig, case_id: str | None, dry_run: bool):
    """Upload observations from an analysis to Loqusdb."""
    click.echo(click.style("----------------- OBSERVATIONS -----------------"))
    try:
        observations_api: MipDNAObservationsAPI | BalsamicObservationsAPI = get_observations_api(
            context=context, case_id=case_id, upload=True
        )
        if dry_run:
            LOG.info(f"Dry run. Would upload observations for {case_id}.")
            return
        observations_api.upload(case_id)
    except CgError:
        LOG.error(f"Could not upload {case_id} to Loqusdb")


@click.command("available-observations")
@OPTION_LOQUSDB_SUPPORTED_WORKFLOW
@DRY_RUN
@click.pass_context
def upload_available_observations_to_loqusdb(
    context: click.Context, workflow: Workflow | None, dry_run: bool
):
    """Uploads the available observations to Loqusdb."""
    click.echo(click.style("----------------- AVAILABLE OBSERVATIONS -----------------"))
    status_db: Store = context.obj.status_db
    cases_to_upload: Query = status_db.observations_to_upload(workflow)
    if not cases_to_upload:
        LOG.error(
            f"There are no available cases to upload to Loqusdb for {workflow} ({datetime.now()})"
        )
        return
    exit_code: int = EXIT_SUCCESS
    for case in cases_to_upload:
        try:
            LOG.info(f"Will upload observations for {case.internal_id}")
            context.invoke(
                upload_observations_to_loqusdb, case_id=case.internal_id, dry_run=dry_run
            )
        except Exception as error:
            LOG.error(f"Error uploading observations for {case.internal_id}: {error}")
            exit_code = EXIT_FAIL
    if exit_code:
        raise click.Abort



from datetime import datetime

import pytest
from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData
from cg.services.illumina.file_parsing.sequencing_times.hiseq_2500_sequencing_times_service import (
    Hiseq2500SequencingTimesService,
)

from cg.services.illumina.file_parsing.sequencing_times.hiseq_x_sequencing_times_service import (
    HiseqXSequencingTimesService,
)
from cg.services.illumina.file_parsing.sequencing_times.novaseq_6000_sequencing_times import (
    Novaseq6000SequencingTimesService,
)
from cg.services.illumina.file_parsing.sequencing_times.novaseq_x_sequencing_times_service import (
    NovaseqXSequencingTimesService,
)
from cg.utils.files import get_source_modified_time_stamp
from cg.utils.time import format_time_from_string, format_time_from_ctime


@pytest.fixture
def novaseq_x_sequencing_times_service() -> NovaseqXSequencingTimesService:
    return NovaseqXSequencingTimesService()


@pytest.fixture
def novaseq_x_sequencing_start_time() -> datetime:
    return format_time_from_string("2023-11-08T16:06:11.5342935+01:00")


@pytest.fixture
def novaseq_x_sequencing_end_time() -> datetime:
    return format_time_from_string("2023-11-09T15:24:13.793134+01:00")


@pytest.fixture
def novaseq_6000_sequencing_times_service() -> Novaseq6000SequencingTimesService:
    return Novaseq6000SequencingTimesService()


@pytest.fixture
def novaseq_6000_sequencing_start_time(
    novaseq_6000_pre_1_5_kits_flow_cell: IlluminaRunDirectoryData,
) -> datetime:
    return novaseq_6000_pre_1_5_kits_flow_cell.sequenced_at


@pytest.fixture
def novaseq_6000_sequencing_end_time(
    novaseq_6000_pre_1_5_kits_flow_cell: IlluminaRunDirectoryData,
) -> datetime:
    int_time: float = get_source_modified_time_stamp(
        novaseq_6000_pre_1_5_kits_flow_cell.get_sequencing_completed_path
    )
    return format_time_from_ctime(int_time)


@pytest.fixture
def hiseq_x_sequencing_times_service() -> HiseqXSequencingTimesService:
    return HiseqXSequencingTimesService()


@pytest.fixture
def hiseq_x_sequencing_start_time(
    hiseq_x_single_index_flow_cell: IlluminaRunDirectoryData,
) -> datetime:
    return hiseq_x_single_index_flow_cell.sequenced_at


@pytest.fixture
def hiseq_x_sequencing_end_time(
    hiseq_x_single_index_flow_cell: IlluminaRunDirectoryData,
) -> datetime:
    int_time: float = get_source_modified_time_stamp(
        hiseq_x_single_index_flow_cell.get_sequencing_completed_path
    )
    return format_time_from_ctime(int_time)


@pytest.fixture
def hiseq_2500_sequencing_times_service() -> Hiseq2500SequencingTimesService:
    return Hiseq2500SequencingTimesService()


@pytest.fixture
def hiseq_2500_sequencing_start_time(
    hiseq_2500_dual_index_flow_cell: IlluminaRunDirectoryData,
) -> datetime:
    return hiseq_2500_dual_index_flow_cell.sequenced_at


@pytest.fixture
def hiseq_2500_sequencing_end_time(
    hiseq_2500_dual_index_flow_cell: IlluminaRunDirectoryData,
) -> datetime:
    return hiseq_2500_dual_index_flow_cell.sequenced_at



"""Tests for the calculations module."""

import math

import pytest

from cg.utils.calculations import (
    divide_by_thousand_with_one_decimal,
    fraction_to_percent,
    multiply_by_million,
)


def test_divide_by_thousand_with_one_decimal():
    # GIVEN a number
    number: float = 1_234.00

    # WHEN dividing by a thousand
    divided: float = divide_by_thousand_with_one_decimal(number)

    # THEN the number should be divided by a thousand and only one decimal should be kept
    assert math.isclose(divided, 1.2, abs_tol=1e-9)


@pytest.mark.parametrize(
    "fraction, expected",
    [
        (0.50, 50.00),
        (0.001, 0.1),
        (2, 2),
    ],
)
def test_fraction_to_percent(fraction: float, expected: float):
    # GIVEN a fraction

    # WHEN converting the fraction to a percentage
    percentage: float = fraction_to_percent(fraction)

    # THEN the fraction should be converted to a percentage
    assert math.isclose(percentage, expected, abs_tol=1e-9)


def test_multiple_by_a_million():
    # GIVEN a number
    number: float = 1.00

    # WHEN multiplying by a million
    multiplied: int = multiply_by_million(number)

    # THEN the number should be multiplied by a million
    assert multiplied == int(number * 1_000_000)



from sqlalchemy.orm import Query

from cg.store.filters.status_customer_filters import (
    filter_customer_by_customer_internal_id,
    filter_customer_by_exclude_customer_internal_id,
)
from cg.store.models import Customer
from cg.store.store import Store


def test_filter_customer_by_customer_id(base_store: Store, customer_id: str):
    """Test return customer by customer internal id."""
    # GIVEN a store containing customers

    # WHEN retrieving a customer
    customer: Customer = filter_customer_by_customer_internal_id(
        customers=base_store._get_query(table=Customer),
        customer_internal_id=customer_id,
    ).first()

    # THEN a customer should be returned
    assert customer

    # THEN the internal id should match the original
    assert customer.internal_id == customer_id


def test_filter_customer_by_exclude_customer_internal_id_exclude_one(
    store_with_cases_and_customers: Store,
):
    """Test that all customers except the one with the excluded internal ID are returned."""
    # GIVEN a store containing multiple customers
    customers_query: Query = store_with_cases_and_customers._get_query(table=Customer)
    exclude_internal_id = customers_query.first().internal_id

    # WHEN filtering customers by excluding a specific customer internal ID
    filtered_customers = filter_customer_by_exclude_customer_internal_id(
        customers=customers_query, exclude_customer_internal_id=exclude_internal_id
    )

    # THEN the query should return all customers except the one with the excluded internal ID
    assert filtered_customers.count() == (customers_query.count() - 1)
    for customer in filtered_customers:
        assert customer.internal_id != exclude_internal_id


def test_filter_customer_by_exclude_customer_internal_id_exclude_nonexistent(
    store_with_cases_and_customers: Store,
):
    """Test that all customers are returned when excluding a nonexistent internal ID."""
    # GIVEN a store containing multiple customers
    customers_query: Query = store_with_cases_and_customers._get_query(table=Customer)
    exclude_internal_id = "nonexistent_id"

    # WHEN filtering customers by excluding a nonexistent customer internal ID
    filtered_customers = filter_customer_by_exclude_customer_internal_id(
        customers=customers_query, exclude_customer_internal_id=exclude_internal_id
    )

    # THEN the query should return all customers
    assert filtered_customers.count() == customers_query.count()



import click

DRY_RUN = click.option(
    "--dry-run",
    is_flag=True,
    default=False,
    help="Runs the command without making any changes",
)

SKIP_CONFIRMATION = click.option(
    "-y",
    "--skip-confirmation",
    "--yes",
    is_flag=True,
    default=False,
    help="Skip confirmation",
)

FORCE = click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Override any warnings or confirmation prompts",
)



from typing import Any

import pytest

from cg.constants.pacbio import (
    CCSAttributeIDs,
    ControlAttributeIDs,
    LoadingAttributesIDs,
    PolymeraseDataAttributeIDs,
    SmrtLinkDatabasesIDs,
)
from cg.services.pacbio.metrics.models import (
    ControlMetrics,
    HiFiMetrics,
    PolymeraseMetrics,
    ProductivityMetrics,
    SmrtlinkDatasetsMetrics,
)


@pytest.fixture
def pac_bio_hifi_metrics() -> HiFiMetrics:
    data: dict[str, Any] = {
        CCSAttributeIDs.NUMBER_OF_READS: 6580977,
        CCSAttributeIDs.TOTAL_NUMBER_OF_BASES: 106192944185,
        CCSAttributeIDs.MEAN_READ_LENGTH: 16136,
        CCSAttributeIDs.MEDIAN_READ_LENGTH: 16205,
        CCSAttributeIDs.READ_LENGTH_N50: 18372,
        CCSAttributeIDs.MEDIAN_ACCURACY: "Q34",
        CCSAttributeIDs.PERCENT_Q30: 0.9318790946286002,
    }
    return HiFiMetrics.model_validate(data, from_attributes=True)


@pytest.fixture
def pac_bio_control_metrics() -> ControlMetrics:
    data: dict[str, Any] = {
        ControlAttributeIDs.NUMBER_OF_READS: 2750,
        ControlAttributeIDs.MEAN_READ_LENGTH: 57730,
        ControlAttributeIDs.PERCENT_MEAN_READ_CONCORDANCE: 0.906334,
        ControlAttributeIDs.PERCENT_MODE_READ_CONCORDANCE: 0.91,
    }
    return ControlMetrics.model_validate(data, from_attributes=True)


@pytest.fixture
def pac_bio_productivity_metrics() -> ProductivityMetrics:
    data: dict[str, Any] = {
        LoadingAttributesIDs.PRODUCTIVE_ZMWS: 25165824,
        LoadingAttributesIDs.P_0: 10012557,
        LoadingAttributesIDs.P_1: 15048838,
        LoadingAttributesIDs.P_2: 104429,
    }
    return ProductivityMetrics.model_validate(data, from_attributes=True)


@pytest.fixture
def pac_bio_polymerase_metrics() -> PolymeraseMetrics:
    data: dict[str, Any] = {
        PolymeraseDataAttributeIDs.MEAN_READ_LENGTH: 82182,
        PolymeraseDataAttributeIDs.READ_LENGTH_N50: 153250,
        PolymeraseDataAttributeIDs.MEAN_LONGEST_SUBREAD_LENGTH: 18466,
        PolymeraseDataAttributeIDs.LONGEST_SUBREAD_LENGTH_N50: 22250,
    }
    return PolymeraseMetrics.model_validate(data, from_attributes=True)


@pytest.fixture
def pac_bio_smrtlink_databases_metrics() -> SmrtlinkDatasetsMetrics:
    data: dict[str, Any] = {
        SmrtLinkDatabasesIDs.BIO_SAMPLE_NAME: "1247014000119",
        SmrtLinkDatabasesIDs.CELL_ID: "EA094834",
        SmrtLinkDatabasesIDs.CELL_INDEX: 0,
        SmrtLinkDatabasesIDs.MOVIE_NAME: "m84202_240522_135641_s1",
        SmrtLinkDatabasesIDs.PATH: "/srv/cg_data/pacbio/r84202_20240522_133539/1_A01/pb_formats/m84202_240522_135641_s1.hifi_reads.consensusreadset.xml",
        SmrtLinkDatabasesIDs.WELL_NAME: "A01",
        SmrtLinkDatabasesIDs.WELL_SAMPLE_NAME: "1247014000119",
    }
    return SmrtlinkDatasetsMetrics.model_validate(data, from_attributes=True)



"""Module for Rnafusion Analysis API."""

import logging
from pathlib import Path

from cg.constants import Workflow
from cg.constants.constants import GenomeVersion, Strandedness
from cg.constants.nf_analysis import MULTIQC_NEXFLOW_CONFIG, RNAFUSION_METRIC_CONDITIONS
from cg.constants.scout import RNAFUSION_CASE_TAGS
from cg.exc import MissingMetrics
from cg.meta.workflow.nf_analysis import NfAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.models.deliverables.metric_deliverables import MetricsBase
from cg.models.rnafusion.rnafusion import RnafusionParameters, RnafusionSampleSheetEntry
from cg.resources import RNAFUSION_BUNDLE_FILENAMES_PATH
from cg.store.models import CaseSample, Sample

LOG = logging.getLogger(__name__)


class RnafusionAnalysisAPI(NfAnalysisAPI):
    """Handles communication between RNAFUSION processes
    and the rest of CG infrastructure."""

    def __init__(
        self,
        config: CGConfig,
        workflow: Workflow = Workflow.RNAFUSION,
    ):
        super().__init__(config=config, workflow=workflow)
        self.root_dir: str = config.rnafusion.root
        self.nfcore_workflow_path: str = config.rnafusion.workflow_path
        self.references: str = config.rnafusion.references
        self.profile: str = config.rnafusion.profile
        self.conda_env: str = config.rnafusion.conda_env
        self.conda_binary: str = config.rnafusion.conda_binary
        self.tower_binary_path: str = config.tower_binary_path
        self.tower_workflow: str = config.rnafusion.tower_workflow
        self.account: str = config.rnafusion.slurm.account
        self.email: str = config.rnafusion.slurm.mail_user
        self.compute_env_base: str = config.rnafusion.compute_env
        self.revision: str = config.rnafusion.revision
        self.nextflow_binary_path: str = config.rnafusion.binary_path

    @property
    def sample_sheet_headers(self) -> list[str]:
        """Headers for sample sheet."""
        return RnafusionSampleSheetEntry.headers()

    @property
    def is_params_appended_to_nextflow_config(self) -> bool:
        """Return True if parameters should be added into the nextflow config file instead of the params file."""
        return False

    @property
    def is_multiple_samples_allowed(self) -> bool:
        """Return whether the analysis supports multiple samples to be linked to the case."""
        return False

    def get_genome_build(self, case_id: str) -> GenomeVersion:
        """Return reference genome for a case. Currently fixed for hg38."""
        return GenomeVersion.HG38

    def get_nextflow_config_content(self, case_id: str) -> str:
        """Return nextflow config content."""
        return MULTIQC_NEXFLOW_CONFIG

    @staticmethod
    def get_bundle_filenames_path() -> Path:
        """Return Rnafusion bundle filenames path."""
        return RNAFUSION_BUNDLE_FILENAMES_PATH

    def get_sample_sheet_content_per_sample(self, case_sample: CaseSample) -> list[list[str]]:
        """Collect and format information required to build a sample sheet for a single sample."""
        fastq_forward_read_paths, fastq_reverse_read_paths = self.get_paired_read_paths(
            sample=case_sample.sample
        )
        sample_sheet_entry = RnafusionSampleSheetEntry(
            name=case_sample.case.internal_id,
            fastq_forward_read_paths=fastq_forward_read_paths,
            fastq_reverse_read_paths=fastq_reverse_read_paths,
            strandedness=Strandedness.REVERSE,
        )
        return sample_sheet_entry.reformat_sample_content()

    def get_workflow_parameters(
        self, case_id: str, genomes_base: Path | None = None
    ) -> RnafusionParameters:
        """Get Rnafusion parameters."""
        return RnafusionParameters(
            cluster_options=f"--qos={self.get_slurm_qos_for_case(case_id=case_id)}",
            genomes_base=genomes_base or self.get_references_path(),
            input=self.get_sample_sheet_path(case_id=case_id),
            outdir=self.get_case_path(case_id=case_id),
            priority=self.account,
        )

    def get_references_path(self, genomes_base: Path | None = None) -> Path:
        if genomes_base:
            return genomes_base.absolute()
        return Path(self.references).absolute()

    def get_multiqc_search_patterns(self, case_id: str) -> dict:
        """Return search patterns for MultiQC for Rnafusion."""
        samples: list[Sample] = self.status_db.get_samples_by_case_id(case_id=case_id)
        search_patterns: dict[str, str] = {case_id: sample.internal_id for sample in samples}
        return search_patterns

    @staticmethod
    def ensure_mandatory_metrics_present(metrics: list[MetricsBase]) -> None:
        """Check that all mandatory metrics are present.
        Raise: MissingMetrics if mandatory metrics are missing."""
        given_metrics: set = {metric.name for metric in metrics}
        mandatory_metrics: set = set(RNAFUSION_METRIC_CONDITIONS.keys())
        LOG.info("Mandatory Metrics Keys:")
        for key in mandatory_metrics:
            LOG.info(key)
        missing_metrics: set = mandatory_metrics.difference(given_metrics)
        if missing_metrics:
            LOG.error(f"Some mandatory metrics are missing: {', '.join(missing_metrics)}")
            raise MissingMetrics()

    def get_workflow_metrics(self, metric_id: str) -> dict:
        return RNAFUSION_METRIC_CONDITIONS

    def get_scout_upload_case_tags(self) -> dict:
        """Return Rnafusion Scout upload case tags."""
        return RNAFUSION_CASE_TAGS



"""Some helper functions for working with files"""

import logging
import os
import shutil
from pathlib import Path

LOG = logging.getLogger(__name__)


def get_file_in_directory(directory: Path, file_name: str) -> Path:
    """Get a file in a directory and subdirectories.
    Raises:
        FileNotFoundError: If the directory does not exist.
    """
    if not directory.is_dir() or not directory.exists():
        raise FileNotFoundError(f"Directory {directory} does not exist")
    for directory_path, _, files in os.walk(directory):
        for file in files:
            if file_name == file:
                path_to_file = Path(directory_path, file)
                return path_to_file
    raise FileNotFoundError(f"File {file_name} not found in {directory}")


def get_files_in_directory_with_pattern(directory: Path, pattern: str) -> list[Path]:
    """Get files with a pattern in a directory and subdirectories.
    Raises:
        FileNotFoundError: If no files with the pattern can be found.
    """
    files_with_pattern: list[Path] = []
    if not directory.is_dir() or not directory.exists():
        raise FileNotFoundError(f"Directory {directory} does not exist")
    for directory_path, _, files in os.walk(directory):
        for file in files:
            if pattern in file:
                files_with_pattern.append(Path(directory_path, file))
    if not files_with_pattern:
        raise FileNotFoundError(f"No files with pattern {pattern} found in {directory}")
    return files_with_pattern


def get_files_matching_pattern(directory: Path, pattern: str) -> list[Path]:
    """Search for all files in a directory that match a pattern."""
    return list(directory.glob(pattern))


def get_all_files_in_dir(base_path: Path) -> set[Path]:
    """Get a set of all files relative to the given base path."""
    return {file.relative_to(base_path) for file in base_path.rglob("*") if file.is_file()}


def rename_file(file_path: Path, renamed_file_path: Path) -> None:
    """Rename the given file path."""
    file_path.rename(renamed_file_path)
    LOG.debug(f"Renamed {file_path} to {renamed_file_path}.")


def is_pattern_in_file_path_name(file_path: Path, pattern: str) -> bool:
    """Check if a pattern is in a file path."""
    return pattern in file_path.name


def get_source_creation_time_stamp(source_path: Path) -> float:
    """
    Return time stamp that a source is created.
    Raises:
        FileNotFoundError if the source does not exist.
    """
    if not source_path.exists():
        raise FileNotFoundError(f"Directory with path {source_path} is not found.")
    return os.stat(source_path).st_ctime


def remove_directory_and_contents(directory_path):
    """
    Delete a directory and its contents.
    Raises OSError when failed to remove directory.
    """
    try:
        shutil.rmtree(directory_path)
        LOG.info(f"Successfully removed the directory and its contents: {directory_path}")
    except Exception as error:
        raise OSError(f"Failed to remove the directory {directory_path} and its contents: {error}")


def get_directories_in_path(path: Path) -> list[Path]:
    """Get all directories for a specified path.
    Raises FileNotFoundError if the path does not exist.
    """
    if not path.exists():
        raise FileNotFoundError(f"Path {path} does not exist.")
    return [entry for entry in path.iterdir() if entry.is_dir()]


def link_or_overwrite_file(src: Path, dst: Path) -> None:
    """Copy a file from src to dst, overwriting the destination if it exists."""
    if dst.exists():
        LOG.warning(f"{dst} already exists. Overwriting with {src}")
        dst.unlink()
    os.link(src=src, dst=dst)
    LOG.debug(f"Linked {src} to {dst}")


def get_all_files_in_directory_tree(directory: Path) -> list[Path]:
    """Get the relative paths of all files in a directory and its subdirectories."""
    files_in_directory: list[Path] = []
    for subdir, _, files in os.walk(directory):
        subdir = Path(subdir).relative_to(directory)
        files_in_directory.extend([Path(subdir, file) for file in files])
    return files_in_directory


def get_source_modified_time_stamp(source_path: Path) -> float:
    """
    Return time stamp that a source is modified. Works for files and directories.
    Raises:
        FileNotFoundError if the source does not exist.
    """
    if not source_path.exists():
        raise FileNotFoundError(f"Directory with path {source_path} is not found.")
    return os.stat(source_path).st_mtime



import logging
from datetime import datetime

from cg.cli.workflow.fastq.base import (
    store_available_fastq_analysis,
    store_fastq_analysis,
)
from cg.constants.constants import CaseActions, Workflow
from cg.store.models import Analysis, Case, Sample


def test_store_fastq_analysis(another_case_id: str, cli_runner, fastq_context, helpers):
    """Test for CLI command creating an analysis object for a fastq case"""
    # GIVEN a fastq context
    helpers.ensure_case(fastq_context.status_db, case_id=another_case_id)
    case_obj: Case = fastq_context.status_db.get_case_by_internal_id(another_case_id)
    assert not case_obj.analyses

    # WHEN a command is run to create an analysis for the case
    cli_runner.invoke(store_fastq_analysis, [another_case_id], obj=fastq_context)

    # THEN the analysis is created
    assert (
        len(
            fastq_context.status_db._get_query(table=Analysis)
            .filter(Analysis.case_id == case_obj.id)
            .all()
        )
        > 0
    )


def test_store_available_fastq_analysis(
    caplog, another_case_id: str, cli_runner, fastq_context, sample_id: str, helpers
):
    """Test for CLI command creating an analysis object for all fastq cases to be delivered"""
    caplog.set_level(logging.INFO)

    # GIVEN a case with no analysis, a sample that has been sequenced and a fastq context
    sample_obj: Sample = fastq_context.status_db.get_sample_by_internal_id(internal_id=sample_id)
    sample_obj.last_sequenced_at = datetime.now()

    # GIVEN a case with no analysis but which is to be analyzed, a sample that has been sequenced and a fastq context
    case_obj: Case = helpers.add_case_with_sample(
        fastq_context.status_db, case_id=another_case_id, sample_id="sample_for_another_case_id"
    )
    assert not case_obj.analyses
    case_obj.data_analysis = Workflow.FASTQ
    case_obj.action = CaseActions.ANALYZE
    case_obj.samples[0].last_sequenced_at = datetime.now()
    case_obj.samples[0].reads = 1

    # WHEN the store_available_fastq_analysis command is invoked
    cli_runner.invoke(store_available_fastq_analysis, ["--dry-run"], obj=fastq_context)

    # THEN the right case should be found and the store_fastq_analysis command should be reached
    assert f"Creating an analysis for case {another_case_id}" in caplog.text



import logging
from pathlib import Path

import click

from cg.apps.demultiplex.sample_sheet.read_sample_sheet import get_samples_from_content
from cg.apps.demultiplex.sample_sheet.sample_models import IlluminaSampleIndexSetting
from cg.apps.demultiplex.sample_sheet.sample_sheet_creator import SampleSheetCreator
from cg.apps.demultiplex.sample_sheet.sample_sheet_validator import SampleSheetValidator
from cg.apps.demultiplex.sample_sheet.utils import (
    delete_sample_sheet_from_housekeeper,
    add_and_include_sample_sheet_path_to_housekeeper,
)
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.apps.lims.sample_sheet import get_flow_cell_samples
from cg.constants.constants import FileFormat
from cg.constants.demultiplexing import SampleSheetBcl2FastqSections, SampleSheetBCLConvertSections
from cg.exc import (
    CgError,
    FlowCellError,
    HousekeeperFileMissingError,
    LimsDataError,
    MissingFilesError,
    SampleSheetContentError,
    SampleSheetFormatError,
)
from cg.io.controller import ReadFile, WriteFile, WriteStream
from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData
from cg.utils.files import get_directories_in_path, link_or_overwrite_file

LOG = logging.getLogger(__name__)


class IlluminaSampleSheetService:
    """Sample Sheet API class."""

    def __init__(self, flow_cell_dir: str, hk_api: HousekeeperAPI, lims_api: LimsAPI) -> None:
        self.flow_cell_runs_dir = Path(flow_cell_dir)
        self.hk_api: HousekeeperAPI = hk_api
        self.lims_api: LimsAPI = lims_api
        self.dry_run: bool = False
        self.force: bool = False
        self.validator = SampleSheetValidator()

    def set_dry_run(self, dry_run: bool) -> None:
        """Set dry run."""
        LOG.debug(f"Set dry run to {dry_run}")
        self.dry_run = dry_run

    def set_force(self, force: bool) -> None:
        """Set force."""
        LOG.debug(f"Set force to {force}")
        self.force = force

    def _get_flow_cell(self, flow_cell_name: str) -> IlluminaRunDirectoryData:
        """
        Return a flow cell given a path.
        Raises:
            SampleSheetError: If the flow cell directory or the data it contains is not valid.
        """
        flow_cell_path: Path = Path(self.flow_cell_runs_dir, flow_cell_name)
        if not flow_cell_path.exists():
            message: str = f"Could not find flow cell {flow_cell_path}"
            LOG.warning(message)
            raise FlowCellError(message)
        flow_cell = IlluminaRunDirectoryData(flow_cell_path)
        return flow_cell

    def validate_sample_sheet(self, sample_sheet_path: Path) -> None:
        """Return the sample sheet path if it exists and if it passes validation.
        Raises:
            MissingFilesError: If the sample sheet does not exist.
            SampleSheetContentError: If the sample sheet content is wrong.
            SampleSheetFormatError: If the sample sheet format is wrong.
        """
        if not (sample_sheet_path and sample_sheet_path.exists()):
            message: str = f"Sample sheet with path {sample_sheet_path} does not exist"
            LOG.error(message)
            raise MissingFilesError(message)
        sample_sheet_content: list[list[str]] = ReadFile.get_content_from_file(
            file_format=FileFormat.CSV, file_path=sample_sheet_path
        )
        self.validator.validate_sample_sheet_from_content(sample_sheet_content)

    @staticmethod
    def _are_necessary_files_in_flow_cell(flow_cell: IlluminaRunDirectoryData) -> bool:
        """Determine if the flow cell has a Run Parameters file and a sample sheet."""
        try:
            flow_cell.run_parameters_path.exists()
        except FlowCellError:
            LOG.error(f"Run parameters file for flow cell {flow_cell.full_name} does not exist")
            return False
        if not flow_cell.sample_sheet_path.exists():
            LOG.error(f"Sample sheet for flow cell {flow_cell.full_name} does not exist")
            return False
        return True

    @staticmethod
    def _replace_sample_header(sample_sheet_content: list[list[str]]) -> list[list[str]]:
        """
        Replace the old sample ID header in the Bcl2Fastq sample sheet content with the BCLConvert
        formatted one.
        Raises:
            SampleSheetFormatError: If the data header is not found in the sample sheet.
        """
        for line in sample_sheet_content:
            if SampleSheetBcl2FastqSections.Data.SAMPLE_INTERNAL_ID_BCL2FASTQ.value in line:
                idx = line.index(
                    SampleSheetBcl2FastqSections.Data.SAMPLE_INTERNAL_ID_BCL2FASTQ.value
                )
                line[idx] = SampleSheetBCLConvertSections.Data.SAMPLE_INTERNAL_ID.value
                return sample_sheet_content
        raise SampleSheetFormatError("Could not find BCL2FASTQ data header in sample sheet")

    def translate_sample_sheet(self, flow_cell_name: str) -> None:
        """Translate a Bcl2Fastq sample sheet to a BCLConvert sample sheet."""
        run_directory_data: IlluminaRunDirectoryData = self._get_flow_cell(flow_cell_name)
        if not self._are_necessary_files_in_flow_cell(run_directory_data):
            raise MissingFilesError("Missing necessary files in run directory for translation")
        original_content: list[list[str]] = ReadFile.get_content_from_file(
            file_format=FileFormat.CSV, file_path=run_directory_data.sample_sheet_path
        )
        content_with_fixed_header: list[list[str]] = self._replace_sample_header(original_content)

        samples: list[IlluminaSampleIndexSetting] = get_samples_from_content(
            sample_sheet_content=content_with_fixed_header
        )
        bcl_convert_creator = SampleSheetCreator(
            run_directory_data=run_directory_data, samples=samples
        )
        new_content = bcl_convert_creator.construct_sample_sheet()
        self.validator.validate_sample_sheet_from_content(new_content)
        if self.dry_run:
            click.echo(
                WriteStream.write_stream_from_content(
                    file_format=FileFormat.CSV, content=new_content
                )
            )
            return
        WriteFile.write_file_from_content(
            content=new_content,
            file_format=FileFormat.CSV,
            file_path=run_directory_data.sample_sheet_path,
        )

    def _use_sample_sheet_from_housekeeper(
        self, run_directory_data: IlluminaRunDirectoryData
    ) -> None:
        """
        Copy the sample sheet from Housekeeper to the flow cell directory if it exists and is valid.
        """
        sample_sheet_path: Path = self.hk_api.get_sample_sheet_path(run_directory_data.id)
        run_directory_data.set_sample_sheet_path_hk(sample_sheet_path)
        self.validate_sample_sheet(sample_sheet_path)

        if self.dry_run:
            LOG.info(
                "Sample sheet from Housekeeper is valid, "
                "would have copied it to sequencing run directory"
            )
            return
        LOG.info("Sample sheet from Housekeeper is valid. Copying it to sequencing run directory")
        link_or_overwrite_file(src=sample_sheet_path, dst=run_directory_data.sample_sheet_path)

    def _use_flow_cell_sample_sheet(self, flow_cell: IlluminaRunDirectoryData) -> None:
        """Use the sample sheet from the flow cell directory if it is valid."""
        self.validate_sample_sheet(flow_cell.sample_sheet_path)
        if self.dry_run:
            LOG.info(
                "Sample sheet from sequencing run directory is valid, "
                "would have added it to Housekeeper"
            )
            return
        LOG.info("Sample sheet from sequencing run directory is valid. Adding it to Housekeeper")
        try:
            delete_sample_sheet_from_housekeeper(flow_cell_id=flow_cell.id, hk_api=self.hk_api)
        except HousekeeperFileMissingError:
            pass
        add_and_include_sample_sheet_path_to_housekeeper(
            flow_cell_directory=flow_cell.path,
            flow_cell_name=flow_cell.id,
            hk_api=self.hk_api,
        )

    def _get_sample_sheet_content(self, flow_cell: IlluminaRunDirectoryData) -> list[list[str]]:
        """Return the sample sheet content for a flow cell.
        Raises:
             LimsDataError: If no samples are found in LIMS for the flow cell.
        """
        lims_samples: list[IlluminaSampleIndexSetting] = list(
            get_flow_cell_samples(
                lims=self.lims_api,
                flow_cell_id=flow_cell.id,
            )
        )
        if not lims_samples:
            message: str = f"Could not find any samples in LIMS for {flow_cell.id}"
            LOG.warning(message)
            raise LimsDataError(message)
        creator = SampleSheetCreator(run_directory_data=flow_cell, samples=lims_samples)
        LOG.info(
            f"Constructing sample sheet for the {flow_cell.sequencer_type} flow cell {flow_cell.id}"
        )
        return creator.construct_sample_sheet()

    def _create_sample_sheet_file(self, flow_cell: IlluminaRunDirectoryData) -> None:
        """Create a valid sample sheet in the flow cell directory and add it to Housekeeper."""
        sample_sheet_content: list[list[str]] = self._get_sample_sheet_content(flow_cell)
        if not self.force:
            self.validator.validate_sample_sheet_from_content(sample_sheet_content)
        LOG.info(f"Writing sample sheet to {flow_cell.sample_sheet_path.resolve()}")
        if self.dry_run:
            click.echo(
                WriteStream.write_stream_from_content(
                    file_format=FileFormat.CSV, content=sample_sheet_content
                )
            )
            return
        WriteFile.write_file_from_content(
            content=sample_sheet_content,
            file_format=FileFormat.CSV,
            file_path=flow_cell.sample_sheet_path,
        )
        try:
            delete_sample_sheet_from_housekeeper(flow_cell_id=flow_cell.id, hk_api=self.hk_api)
        except HousekeeperFileMissingError:
            pass
        add_and_include_sample_sheet_path_to_housekeeper(
            flow_cell_directory=flow_cell.path, flow_cell_name=flow_cell.id, hk_api=self.hk_api
        )

    def get_or_create_sample_sheet(self, flow_cell_name: str) -> None:
        """
        Ensure that a valid sample sheet is present in the flow cell directory by fetching it from
        housekeeper or creating it if there is not a valid sample sheet.
        """
        flow_cell: IlluminaRunDirectoryData = self._get_flow_cell(flow_cell_name)
        LOG.debug(f"Fetching and validating sample sheet for {flow_cell_name} from Housekeeper")
        try:
            self._use_sample_sheet_from_housekeeper(flow_cell)
            return
        except SampleSheetContentError:
            LOG.warning(
                f"Validation failed for {flow_cell.get_sample_sheet_path_hk()}. "
                "Possibly a manually modified sample sheet. Sample sheet will not be re-generated."
            )
            return
        except CgError:
            LOG.warning(
                "Sample sheet from Housekeeper is not correctly formatted or does not exist, "
                "trying sample sheet in sequencing directory"
            )
        try:
            self._use_flow_cell_sample_sheet(flow_cell)
            return
        except SampleSheetContentError:
            LOG.warning(
                f"Validation failed for {flow_cell.sample_sheet_path}. "
                "Possibly manually modified sample sheet. Sample sheet will not be re-generated."
            )
            return
        except CgError:
            LOG.warning(
                "Sample sheet from sequencing directory is not correctly formatted or does not "
                "exist, creating new sample sheet"
            )
        self._create_sample_sheet_file(flow_cell)

    def get_or_create_all_sample_sheets(self):
        """Ensure that a valid sample sheet is present in all flow cell directories."""
        for flow_cell_dir in get_directories_in_path(self.flow_cell_runs_dir):
            LOG.info(f"Getting a valid sample sheet for flow cell {flow_cell_dir.name}")
            try:
                self.get_or_create_sample_sheet(flow_cell_dir.name)
            except Exception as error:
                LOG.error(f"Could not create sample sheet for {flow_cell_dir.name}: {error}")
                continue



from datetime import datetime
import os
from pathlib import Path

from cg.exc import MissingAnalysisRunDirectory


def get_project_directory_date(dir_name: str) -> datetime:
    # Assumes format like <project_id>_year.month.day_hour.minute.second
    _, date, time = dir_name.split("_")
    date_time = f"{date}_{time}"
    return datetime.strptime(date_time, "%Y.%m.%d_%H.%M.%S")


def get_project_directories(project_id: str, directory: Path) -> list[str]:
    return [sub_dir for sub_dir in os.listdir(directory) if sub_dir.startswith(project_id)]


def sort_project_directories_by_date(project_directories: list[str]) -> list[str]:
    return sorted(project_directories, key=get_project_directory_date, reverse=True)


def get_most_recent_project_directory(project_id: str, directory: Path) -> Path:
    project_directories: list[str] = get_project_directories(
        project_id=project_id, directory=directory
    )
    sorted_project_directories: list[str] = sort_project_directories_by_date(project_directories)

    if not sorted_project_directories:
        raise MissingAnalysisRunDirectory(f"No analysis directory found for project {project_id}.")

    return Path(directory, sorted_project_directories[0])



"""Module to test the collect sequencing times class."""

import pytest
from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData

from cg.services.illumina.file_parsing.sequencing_times.collect_sequencing_times import (
    CollectSequencingTimes,
)
from cg.services.illumina.file_parsing.sequencing_times.hiseq_2500_sequencing_times_service import (
    Hiseq2500SequencingTimesService,
)
from cg.services.illumina.file_parsing.sequencing_times.hiseq_x_sequencing_times_service import (
    HiseqXSequencingTimesService,
)
from cg.services.illumina.file_parsing.sequencing_times.novaseq_6000_sequencing_times import (
    Novaseq6000SequencingTimesService,
)
from cg.services.illumina.file_parsing.sequencing_times.novaseq_x_sequencing_times_service import (
    NovaseqXSequencingTimesService,
)
from cg.services.illumina.file_parsing.sequencing_times.sequencing_time_service import (
    SequencingTimesService,
)


@pytest.mark.parametrize(
    "run_dir_data, expected_times_service",
    [
        ("novaseq_x_flow_cell", NovaseqXSequencingTimesService),
        ("novaseq_6000_pre_1_5_kits_flow_cell", Novaseq6000SequencingTimesService),
        ("hiseq_x_single_index_flow_cell", HiseqXSequencingTimesService),
        ("hiseq_2500_dual_index_flow_cell", Hiseq2500SequencingTimesService),
    ],
)
def test_collect_sequencing_time_get_service(
    run_dir_data: str, expected_times_service: SequencingTimesService, request
):
    # GIVEN run directory data and a collect sequencing time class
    sequencing_time_collector = CollectSequencingTimes()
    run_dir_data: IlluminaRunDirectoryData = request.getfixturevalue(run_dir_data)

    # WHEN retrieving the time service
    time_service: SequencingTimesService = sequencing_time_collector._get_sequencing_times_service(
        run_dir_data
    )

    # THEN the correct time service is returned
    assert time_service == expected_times_service



import shutil
from pathlib import Path
from typing import Callable

import pytest
from _pytest.fixtures import FixtureRequest

from cg.constants.pacbio import PacBioDirsAndFiles
from cg.exc import PacBioMetricsParsingError
from cg.services.pacbio.metrics.models import BaseMetrics
from cg.services.pacbio.metrics.utils import (
    parse_control_metrics,
    parse_dataset_metrics,
    parse_hifi_metrics,
    parse_polymerase_metrics,
    parse_productivity_metrics,
)


@pytest.mark.parametrize(
    "parsing_function, expected_metrics_fixture",
    [
        (parse_control_metrics, "pac_bio_control_metrics"),
        (parse_dataset_metrics, "pac_bio_smrtlink_databases_metrics"),
        (parse_hifi_metrics, "pac_bio_hifi_metrics"),
        (parse_productivity_metrics, "pac_bio_productivity_metrics"),
        (parse_polymerase_metrics, "pac_bio_polymerase_metrics"),
    ],
    ids=["Control", "Smrtlink-Dataset", "Hi-Fi", "Productivity", "Polymerase"],
)
def test_parse_dataset_metrics(
    parsing_function: Callable,
    expected_metrics_fixture: str,
    pac_bio_run_statistics_dir: Path,
    request: FixtureRequest,
):
    """Test the parsing of all PacBio metric files."""
    # GIVEN a metrics file to parse in an existing statistics directory

    # WHEN parsing the SMRTlink datasets metrics
    parsed_metrics: BaseMetrics = parsing_function(report_dir=pac_bio_run_statistics_dir)

    # THEN the parsed metrics are the expected ones
    expected_metrics: BaseMetrics = request.getfixturevalue(expected_metrics_fixture)
    assert parsed_metrics == expected_metrics


@pytest.mark.parametrize(
    "parsing_function",
    [
        parse_control_metrics,
        parse_dataset_metrics,
        parse_hifi_metrics,
        parse_productivity_metrics,
        parse_polymerase_metrics,
    ],
    ids=["Control", "Smrtlink-Dataset", "Hi-Fi", "Productivity", "Polymerase"],
)
def test_parse_dataset_metrics_missing_file(tmp_path: Path, parsing_function: Callable):
    """Test the parsing of all PacBio metric files with a missing file raises an error."""
    # GIVEN a directory with no metrics files to parse

    # WHEN parsing the metrics
    with pytest.raises(PacBioMetricsParsingError) as error:
        # THEN a PacBioMetricsParsingError is raised
        parse_dataset_metrics(report_dir=tmp_path)
    assert "Could not find the metrics file" in str(error.value)
    assert tmp_path.as_posix() in str(error.value)


@pytest.mark.parametrize(
    "parsing_function, metrics_file_name",
    [
        (parse_control_metrics, PacBioDirsAndFiles.CONTROL_REPORT),
        (parse_dataset_metrics, PacBioDirsAndFiles.SMRTLINK_DATASETS_REPORT),
        (parse_hifi_metrics, PacBioDirsAndFiles.BASECALLING_REPORT),
        (parse_productivity_metrics, PacBioDirsAndFiles.LOADING_REPORT),
        (parse_polymerase_metrics, PacBioDirsAndFiles.RAW_DATA_REPORT),
    ],
    ids=["Control", "Smrtlink-Dataset", "Hi-Fi", "Productivity", "Polymerase"],
)
def test_parse_dataset_metrics_validation_error(
    tmp_path: Path,
    parsing_function: Callable,
    metrics_file_name: str,
    pac_bio_wrong_metrics_file: Path,
):
    """Test the parsing of all PacBio metric files with an unexpected error raises an error."""
    # GIVEN a tmp statistics directory with a metrics file without the expected structure
    shutil.copyfile(src=pac_bio_wrong_metrics_file, dst=Path(tmp_path, metrics_file_name))
    assert Path(tmp_path, metrics_file_name).exists()

    # WHEN parsing the metrics
    with pytest.raises(PacBioMetricsParsingError) as error:
        # THEN a PacBioMetricsParsingError is raised
        parsing_function(report_dir=tmp_path)
    assert "An error occurred while parsing the metrics" in str(error.value)



"""Tests for the sequencing times run services."""

from datetime import datetime

import pytest
from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData

from cg.services.illumina.file_parsing.sequencing_times.sequencing_time_service import (
    SequencingTimesService,
)


@pytest.mark.parametrize(
    "run_dir_data, time_service, expected_time",
    [
        (
            "novaseq_x_flow_cell",
            "novaseq_x_sequencing_times_service",
            "novaseq_x_sequencing_start_time",
        ),
        (
            "novaseq_6000_pre_1_5_kits_flow_cell",
            "novaseq_6000_sequencing_times_service",
            "novaseq_6000_sequencing_start_time",
        ),
        (
            "hiseq_x_single_index_flow_cell",
            "hiseq_x_sequencing_times_service",
            "hiseq_x_sequencing_start_time",
        ),
        (
            "hiseq_2500_dual_index_flow_cell",
            "hiseq_2500_sequencing_times_service",
            "hiseq_2500_sequencing_start_time",
        ),
    ],
)
def test_get_sequencing_start_time(
    run_dir_data: str,
    time_service: str,
    expected_time: str,
    request,
):
    # GIVEN run directory data and a time service
    run_dir_data: IlluminaRunDirectoryData = request.getfixturevalue(run_dir_data)
    time_service: SequencingTimesService = request.getfixturevalue(time_service)
    expected_time: datetime = request.getfixturevalue(expected_time)

    # WHEN retrieving the start time
    start_time: datetime = time_service.get_start_time(run_dir_data)

    # THEN the expected start time is returned
    assert start_time == expected_time


@pytest.mark.parametrize(
    "run_dir_data, time_service, expected_time",
    [
        (
            "novaseq_x_flow_cell",
            "novaseq_x_sequencing_times_service",
            "novaseq_x_sequencing_end_time",
        ),
        (
            "novaseq_6000_pre_1_5_kits_flow_cell",
            "novaseq_6000_sequencing_times_service",
            "novaseq_6000_sequencing_end_time",
        ),
        (
            "hiseq_x_single_index_flow_cell",
            "hiseq_x_sequencing_times_service",
            "hiseq_x_sequencing_end_time",
        ),
        (
            "hiseq_2500_dual_index_flow_cell",
            "hiseq_2500_sequencing_times_service",
            "hiseq_2500_sequencing_end_time",
        ),
    ],
)
def test_get_sequencing_end_time(
    run_dir_data: str,
    time_service: str,
    expected_time: str,
    request,
):
    # GIVEN run directory data and a time service
    run_dir_data: IlluminaRunDirectoryData = request.getfixturevalue(run_dir_data)
    time_service: SequencingTimesService = request.getfixturevalue(time_service)
    expected_time: datetime = request.getfixturevalue(expected_time)

    # WHEN retrieving the start time
    end_time: datetime = time_service.get_end_time(run_dir_data)

    # THEN the expected start time is returned
    assert end_time == expected_time



import re
from datetime import datetime
from enum import Enum
from typing import Annotated

from sqlalchemy import (
    BLOB,
    DECIMAL,
    VARCHAR,
    BigInteger,
    Column,
    ForeignKey,
    Numeric,
    String,
    Table,
)
from sqlalchemy import Text as SLQText
from sqlalchemy import UniqueConstraint, orm, types
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column
from sqlalchemy.orm.attributes import InstrumentedAttribute

from cg.constants import DataDelivery, Priority, SequencingRunDataAvailability, Workflow
from cg.constants.archiving import PDC_ARCHIVE_LOCATION
from cg.constants.constants import (
    CaseActions,
    ControlOptions,
    PrepCategory,
    SequencingQCStatus,
    SexOptions,
    StatusOptions,
)
from cg.constants.devices import DeviceType
from cg.constants.priority import SlurmQos
from cg.constants.symbols import EMPTY_STRING

BigInt = Annotated[int, None]
Blob = Annotated[bytes, None]
Decimal = Annotated[float, None]
Num_6_2 = Annotated[float, 6]
Str32 = Annotated[str, 32]
Str64 = Annotated[str, 64]
Str128 = Annotated[str, 128]
Str255 = Annotated[str, 255]
Str256 = Annotated[str, 256]
Text = Annotated[str, None]
VarChar128 = Annotated[str, 128]

PrimaryKeyInt = Annotated[int, mapped_column(primary_key=True)]
UniqueStr = Annotated[str, mapped_column(String(32), unique=True)]
UniqueStr64 = Annotated[str, mapped_column(String(64), unique=True)]


class Base(DeclarativeBase):
    type_annotation_map = {
        BigInt: BigInteger,
        Blob: BLOB,
        Decimal: DECIMAL,
        Num_6_2: Numeric(6, 2),
        Str32: String(32),
        Str64: String(64),
        Str128: String(128),
        Str255: String(255),
        Str256: String(256),
        Text: SLQText,
        VarChar128: VARCHAR(128),
    }


def to_dict(model_instance):
    def serialize_value(value):
        if isinstance(value, InstrumentedAttribute):
            return None
        if isinstance(value, Enum):
            return value.name
        return value

    if hasattr(model_instance, "__table__"):
        return {
            column.name: serialize_value(getattr(model_instance, column.name))
            for column in model_instance.__table__.columns
        }
    return {}


customer_user = Table(
    "customer_user",
    Base.metadata,
    Column("customer_id", types.Integer, ForeignKey("customer.id"), nullable=False),
    Column("user_id", types.Integer, ForeignKey("user.id"), nullable=False),
    UniqueConstraint("customer_id", "user_id", name="_customer_user_uc"),
)

customer_collaboration = Table(
    "customer_collaboration",
    Base.metadata,
    Column("customer_id", types.Integer, ForeignKey("customer.id"), nullable=False),
    Column("collaboration_id", types.Integer, ForeignKey("collaboration.id"), nullable=False),
    UniqueConstraint("customer_id", "collaboration_id", name="_customer_collaboration_uc"),
)

order_case = Table(
    "order_case",
    Base.metadata,
    Column("order_id", ForeignKey("order.id", ondelete="CASCADE"), nullable=False),
    Column("case_id", ForeignKey("case.id", ondelete="CASCADE"), nullable=False),
    UniqueConstraint("order_id", "case_id", name="_order_case_uc"),
)


class PriorityMixin:
    @property
    def priority_human(self) -> str:
        """Humanized priority for sample."""
        return self.priority.name

    @priority_human.setter
    def priority_human(self, priority: str) -> None:
        self.priority: Priority = Priority[priority]

    @property
    def priority_int(self) -> int:
        """Priority as integer for sample."""
        return self.priority.value

    @priority_int.setter
    def priority_int(self, priority_int: int) -> None:
        self.priority: Priority = Priority(priority_int)

    @property
    def high_priority(self):
        """Has high priority?"""
        return self.priority_int > 1

    @property
    def low_priority(self):
        """Has low priority?"""
        return self.priority_int < 1


class Application(Base):
    __tablename__ = "application"

    id: Mapped[PrimaryKeyInt]

    tag: Mapped[UniqueStr]
    prep_category: Mapped[str] = mapped_column(
        types.Enum(*(category.value for category in PrepCategory))
    )
    is_external: Mapped[bool] = mapped_column(default=False)
    description: Mapped[Str256]
    is_accredited: Mapped[bool]

    turnaround_time: Mapped[int | None]
    minimum_order: Mapped[int | None] = mapped_column(default=1)
    sequencing_depth: Mapped[int | None]
    min_sequencing_depth: Mapped[int] = mapped_column(default=0)
    target_reads: Mapped[BigInt | None] = mapped_column(default=0)
    percent_reads_guaranteed: Mapped[int]
    sample_amount: Mapped[int | None]
    sample_volume: Mapped[Text | None]
    sample_concentration: Mapped[Text | None]
    sample_concentration_minimum: Mapped[Decimal | None]
    sample_concentration_maximum: Mapped[Decimal | None]
    sample_concentration_minimum_cfdna: Mapped[Decimal | None]
    sample_concentration_maximum_cfdna: Mapped[Decimal | None]
    priority_processing: Mapped[bool] = mapped_column(default=False)
    details: Mapped[Text | None]
    limitations: Mapped[Text | None]
    percent_kth: Mapped[int]
    comment: Mapped[Text | None]
    is_archived: Mapped[bool] = mapped_column(default=False)

    created_at: Mapped[datetime | None] = mapped_column(default=datetime.now)
    updated_at: Mapped[datetime | None] = mapped_column(onupdate=datetime.now)

    versions: Mapped[list["ApplicationVersion"]] = orm.relationship(
        order_by="ApplicationVersion.version", back_populates="application"
    )
    pipeline_limitations: Mapped[list["ApplicationLimitations"]] = orm.relationship(
        back_populates="application"
    )

    def __str__(self) -> str:
        return self.tag

    @property
    def reduced_price(self):
        return self.tag.startswith("WGT") or self.tag.startswith("EXT")

    @property
    def expected_reads(self):
        return self.target_reads * self.percent_reads_guaranteed / 100

    @property
    def analysis_type(self) -> str:
        if self.prep_category == PrepCategory.WHOLE_TRANSCRIPTOME_SEQUENCING.value:
            return PrepCategory.WHOLE_TRANSCRIPTOME_SEQUENCING.value

        return (
            PrepCategory.WHOLE_GENOME_SEQUENCING.value
            if self.prep_category == PrepCategory.WHOLE_GENOME_SEQUENCING.value
            else PrepCategory.WHOLE_EXOME_SEQUENCING.value
        )

    def to_dict(self):
        return to_dict(model_instance=self)


class ApplicationVersion(Base):
    __tablename__ = "application_version"
    __table_args__ = (UniqueConstraint("application_id", "version", name="_app_version_uc"),)

    id: Mapped[PrimaryKeyInt]
    version: Mapped[int]

    valid_from: Mapped[datetime | None] = mapped_column(default=datetime.now)
    price_standard: Mapped[int | None]
    price_priority: Mapped[int | None]
    price_express: Mapped[int | None]
    price_research: Mapped[int | None]
    price_clinical_trials: Mapped[int | None]
    comment: Mapped[Text | None]

    created_at: Mapped[datetime | None] = mapped_column(default=datetime.now)
    updated_at: Mapped[datetime | None] = mapped_column(onupdate=datetime.now)
    application_id: Mapped[int] = mapped_column(ForeignKey(Application.id))

    application: Mapped[Application] = orm.relationship(back_populates="versions")

    def __str__(self) -> str:
        return f"{self.application.tag} ({self.version})"

    def to_dict(self, application: bool = True):
        """Represent as dictionary"""
        data = to_dict(model_instance=self)
        if application:
            data["application"] = self.application.to_dict()
        return data


class ApplicationLimitations(Base):
    __tablename__ = "application_limitations"

    id: Mapped[PrimaryKeyInt]
    application_id: Mapped[int] = mapped_column(ForeignKey(Application.id))
    workflow: Mapped[str] = mapped_column(types.Enum(*(workflow.value for workflow in Workflow)))
    limitations: Mapped[Text | None]
    comment: Mapped[Text | None]
    created_at: Mapped[datetime | None] = mapped_column(default=datetime.now)
    updated_at: Mapped[datetime | None] = mapped_column(onupdate=datetime.now)

    application: Mapped[Application] = orm.relationship(back_populates="pipeline_limitations")

    def __str__(self):
        return f"{self.application.tag} – {self.workflow}"

    def to_dict(self):
        return to_dict(model_instance=self)


class Analysis(Base):
    __tablename__ = "analysis"

    id: Mapped[PrimaryKeyInt]
    workflow: Mapped[str | None] = mapped_column(
        types.Enum(*(workflow.value for workflow in Workflow))
    )
    workflow_version: Mapped[Str32 | None]
    started_at: Mapped[datetime | None]
    completed_at: Mapped[datetime | None]
    delivery_report_created_at: Mapped[datetime | None]
    upload_started_at: Mapped[datetime | None]
    uploaded_at: Mapped[datetime | None]
    cleaned_at: Mapped[datetime | None]
    # primary analysis is the one originally delivered to the customer
    is_primary: Mapped[bool | None] = mapped_column(default=False)
    created_at: Mapped[datetime] = mapped_column(default=datetime.now)
    comment: Mapped[Text | None]
    case_id: Mapped[int] = mapped_column(ForeignKey("case.id", ondelete="CASCADE"))
    case: Mapped["Case"] = orm.relationship(back_populates="analyses")

    def __str__(self):
        return f"{self.case.internal_id} | {self.completed_at.date()}"

    def to_dict(self, family: bool = True):
        """Represent as dictionary"""
        data = to_dict(model_instance=self)
        if family:
            data["family"] = self.case.to_dict()
        return data


class Bed(Base):
    """Model for bed target captures"""

    __tablename__ = "bed"
    id: Mapped[PrimaryKeyInt]
    name: Mapped[UniqueStr]
    comment: Mapped[Text | None]
    is_archived: Mapped[bool] = mapped_column(default=False)
    created_at: Mapped[datetime | None] = mapped_column(default=datetime.now)
    updated_at: Mapped[datetime | None] = mapped_column(onupdate=datetime.now)

    versions: Mapped[list["BedVersion"]] = orm.relationship(
        order_by="BedVersion.version", back_populates="bed"
    )

    def __str__(self) -> str:
        return self.name

    def to_dict(self):
        return to_dict(model_instance=self)


class BedVersion(Base):
    """Model for bed target captures versions"""

    __tablename__ = "bed_version"
    __table_args__ = (UniqueConstraint("bed_id", "version", name="_app_version_uc"),)

    id: Mapped[PrimaryKeyInt]
    shortname: Mapped[Str64 | None]
    version: Mapped[int]
    filename: Mapped[Str256]
    checksum: Mapped[Str32 | None]
    panel_size: Mapped[int | None]
    genome_version: Mapped[Str32 | None]
    designer: Mapped[Str256 | None]
    comment: Mapped[Text | None]
    created_at: Mapped[datetime | None] = mapped_column(default=datetime.now)
    updated_at: Mapped[datetime | None] = mapped_column(onupdate=datetime.now)
    bed_id: Mapped[int] = mapped_column(ForeignKey(Bed.id))

    bed: Mapped[Bed] = orm.relationship(back_populates="versions")

    def __str__(self) -> str:
        return f"{self.bed.name} ({self.version})"

    def to_dict(self, bed: bool = True):
        """Represent as dictionary"""
        data = to_dict(model_instance=self)
        if bed:
            data["bed"] = self.bed.to_dict()
        return data


class Customer(Base):
    __tablename__ = "customer"
    agreement_date: Mapped[datetime | None]
    agreement_registration: Mapped[Str32 | None]
    comment: Mapped[Text | None]
    id: Mapped[PrimaryKeyInt]
    internal_id: Mapped[UniqueStr]
    invoice_address: Mapped[Text]
    invoice_reference: Mapped[Str32]
    is_trusted: Mapped[bool] = mapped_column(default=False)
    lab_contact_id: Mapped[int | None] = mapped_column(ForeignKey("user.id"))
    lab_contact: Mapped["User | None"] = orm.relationship(foreign_keys=[lab_contact_id])
    loqus_upload: Mapped[bool] = mapped_column(default=False)
    name: Mapped[Str128]
    organisation_number: Mapped[Str32 | None]
    priority: Mapped[str | None] = mapped_column(types.Enum("diagnostic", "research"))
    project_account_ki: Mapped[Str32 | None]
    project_account_kth: Mapped[Str32 | None]
    return_samples: Mapped[bool] = mapped_column(default=False)
    scout_access: Mapped[bool] = mapped_column(default=False)
    uppmax_account: Mapped[Str32 | None]
    data_archive_location: Mapped[Str32] = mapped_column(default=PDC_ARCHIVE_LOCATION)
    is_clinical: Mapped[bool] = mapped_column(default=False)

    collaborations: Mapped[list["Collaboration"]] = orm.relationship(
        secondary="customer_collaboration", back_populates="customers"
    )

    delivery_contact_id: Mapped[int | None] = mapped_column(ForeignKey("user.id"))
    delivery_contact: Mapped["User | None"] = orm.relationship(foreign_keys=[delivery_contact_id])
    invoice_contact_id: Mapped[int | None] = mapped_column(ForeignKey("user.id"))
    invoice_contact: Mapped["User | None"] = orm.relationship(foreign_keys=[invoice_contact_id])
    primary_contact_id: Mapped[int | None] = mapped_column(ForeignKey("user.id"))
    primary_contact: Mapped["User | None"] = orm.relationship(foreign_keys=[primary_contact_id])

    panels: Mapped[list["Panel"]] = orm.relationship(back_populates="customer")
    users: Mapped[list["User"]] = orm.relationship(
        secondary=customer_user, back_populates="customers"
    )

    def __str__(self) -> str:
        return f"{self.internal_id} ({self.name})"

    @property
    def collaborators(self) -> set["Customer"]:
        """All customers that the current customer collaborates with (including itself)"""
        customers = {
            customer
            for collaboration in self.collaborations
            for customer in collaboration.customers
        }
        customers.add(self)
        return customers

    def to_dict(self):
        return to_dict(model_instance=self)


class Collaboration(Base):
    __tablename__ = "collaboration"
    id: Mapped[PrimaryKeyInt]
    internal_id: Mapped[Str32] = mapped_column(unique=True)
    name: Mapped[Str128]
    customers: Mapped[list[Customer]] = orm.relationship(
        secondary="customer_collaboration", back_populates="collaborations"
    )

    def __str__(self) -> str:
        return f"{self.internal_id} ({self.name})"

    def to_dict(self):
        """Represent as dictionary"""
        return {
            "customers": [customer.internal_id for customer in self.customers],
            "id": self.id,
            "name": self.name,
            "internal_id": self.internal_id,
        }


class Case(Base, PriorityMixin):
    __tablename__ = "case"
    __table_args__ = (UniqueConstraint("customer_id", "name", name="_customer_name_uc"),)

    action: Mapped[str | None] = mapped_column(
        types.Enum(*(action.value for action in CaseActions))
    )
    _cohorts: Mapped[Text | None]
    comment: Mapped[Text | None]
    created_at: Mapped[datetime | None] = mapped_column(default=datetime.now)
    customer_id: Mapped[int] = mapped_column(ForeignKey("customer.id", ondelete="CASCADE"))
    customer: Mapped["Customer"] = orm.relationship(foreign_keys=[customer_id])
    data_analysis: Mapped[str | None] = mapped_column(
        types.Enum(*(workflow.value for workflow in Workflow))
    )
    data_delivery: Mapped[str | None] = mapped_column(
        types.Enum(*(delivery.value for delivery in DataDelivery))
    )
    id: Mapped[PrimaryKeyInt]
    internal_id: Mapped[UniqueStr]
    is_compressible: Mapped[bool] = mapped_column(default=True)
    name: Mapped[Str128]
    ordered_at: Mapped[datetime | None] = mapped_column(default=datetime.now)
    _panels: Mapped[Text | None]

    priority: Mapped[Priority] = mapped_column(
        default=Priority.standard,
    )

    aggregated_sequencing_qc: Mapped[SequencingQCStatus] = mapped_column(
        types.Enum(SequencingQCStatus), default=SequencingQCStatus.PENDING
    )
    synopsis: Mapped[Text | None]
    tickets: Mapped[VarChar128 | None]

    analyses: Mapped[list[Analysis]] = orm.relationship(
        back_populates="case", order_by="-Analysis.completed_at"
    )
    links: Mapped[list["CaseSample"]] = orm.relationship(back_populates="case")
    orders: Mapped[list["Order"]] = orm.relationship(secondary=order_case, back_populates="cases")

    @property
    def cohorts(self) -> list[str]:
        """Return a list of cohorts."""
        return self._cohorts.split(",") if self._cohorts else []

    @cohorts.setter
    def cohorts(self, cohort_list: list[str]):
        self._cohorts = ",".join(cohort_list) if cohort_list else None

    @property
    def panels(self) -> list[str]:
        """Return a list of panels."""
        return self._panels.split(",") if self._panels else []

    @panels.setter
    def panels(self, panel_list: list[str]):
        self._panels = ",".join(panel_list) if panel_list else None

    @property
    def latest_ticket(self) -> str | None:
        """Returns the last ticket the family was ordered in"""
        return self.tickets.split(sep=",")[-1] if self.tickets else None

    @property
    def latest_order(self) -> "Order":
        """Returns the latest order this case was included in."""
        sorted_orders: list[Order] = sorted(
            self.orders, key=lambda order: order.order_date, reverse=True
        )
        return sorted_orders[0]

    @property
    def latest_analyzed(self) -> datetime | None:
        return self.analyses[0].completed_at if self.analyses else None

    @property
    def latest_sequenced(self) -> datetime | None:
        sequenced_dates = []
        for link in self.links:
            if link.sample.application_version.application.is_external:
                sequenced_dates.append(link.sample.ordered_at)
            elif link.sample.last_sequenced_at:
                sequenced_dates.append(link.sample.last_sequenced_at)
        return max(sequenced_dates, default=None)

    @property
    def are_all_samples_sequenced(self) -> bool:
        return all([link.sample.last_sequenced_at for link in self.links])

    def __str__(self) -> str:
        return f"{self.internal_id} ({self.name})"

    @property
    def samples(self) -> list["Sample"]:
        """Return case samples."""
        return [link.sample for link in self.links]

    @property
    def sample_ids(self) -> list[str]:
        """Return a list of internal ids of the case samples."""
        return [sample.internal_id for sample in self.samples]

    @property
    def tumour_samples(self) -> list["Sample"]:
        """Return tumour samples."""
        return [link.sample for link in self.links if link.sample.is_tumour]

    @property
    def non_tumour_samples(self) -> list["Sample"]:
        """Return non-tumour samples."""
        return [link.sample for link in self.links if not link.sample.is_tumour]

    @property
    def loqusdb_uploaded_samples(self) -> list["Sample"]:
        """Return uploaded samples to Loqusdb."""
        return [link.sample for link in self.links if link.sample.loqusdb_id]

    @property
    def is_uploaded(self) -> bool:
        """Returns True if the latest connected analysis has been uploaded."""
        return self.analyses and self.analyses[0].uploaded_at

    @property
    def slurm_priority(self) -> SlurmQos:
        case_priority: str = self.priority
        slurm_priority: str = Priority.priority_to_slurm_qos().get(case_priority)
        return SlurmQos(slurm_priority)

    def get_delivery_arguments(self) -> set[str]:
        """Translates the case data_delivery field to workflow specific arguments."""
        delivery_arguments: set[str] = set()
        requested_deliveries: list[str] = re.split("[-_]", self.data_delivery)
        delivery_per_workflow_map: dict[str, str] = {
            DataDelivery.FASTQ: Workflow.FASTQ,
            DataDelivery.ANALYSIS_FILES: self.data_analysis,
        }
        for data_delivery, workflow in delivery_per_workflow_map.items():
            if data_delivery in requested_deliveries:
                delivery_arguments.add(workflow)
        return delivery_arguments

    def to_dict(self, links: bool = False, analyses: bool = False) -> dict:
        """Represent as dictionary."""
        data = to_dict(model_instance=self)
        data["panels"] = self.panels
        data["priority"] = self.priority_human
        data["customer"] = self.customer.to_dict()
        if links:
            data["links"] = [link_obj.to_dict(samples=True) for link_obj in self.links]
        if analyses:
            data["analyses"] = [
                analysis_obj.to_dict(family=False) for analysis_obj in self.analyses
            ]
        return data


class CaseSample(Base):
    __tablename__ = "case_sample"
    __table_args__ = (UniqueConstraint("case_id", "sample_id", name="_case_sample_uc"),)

    id: Mapped[PrimaryKeyInt]
    case_id: Mapped[str] = mapped_column(ForeignKey("case.id", ondelete="CASCADE"))
    sample_id: Mapped[int] = mapped_column(ForeignKey("sample.id", ondelete="CASCADE"))
    status: Mapped[str] = mapped_column(
        types.Enum(*(status.value for status in StatusOptions)), default=StatusOptions.UNKNOWN.value
    )

    created_at: Mapped[datetime | None]
    updated_at: Mapped[datetime | None]

    mother_id: Mapped[int | None] = mapped_column(ForeignKey("sample.id"))
    father_id: Mapped[int | None] = mapped_column(ForeignKey("sample.id"))

    case: Mapped[Case] = orm.relationship(back_populates="links")
    sample: Mapped["Sample"] = orm.relationship(foreign_keys=[sample_id], back_populates="links")
    mother: Mapped["Sample | None"] = orm.relationship(
        foreign_keys=[mother_id], back_populates="mother_links"
    )
    father: Mapped["Sample | None"] = orm.relationship(
        foreign_keys=[father_id], back_populates="father_links"
    )

    def to_dict(self, parents: bool = False, samples: bool = False, family: bool = False) -> dict:
        """Represent as dictionary"""
        data = to_dict(model_instance=self)
        if samples:
            data["sample"] = self.sample.to_dict()
            data["mother"] = self.mother.to_dict() if self.mother else None
            data["father"] = self.father.to_dict() if self.father else None
        elif parents:
            data["mother"] = self.mother.to_dict() if self.mother else None
            data["father"] = self.father.to_dict() if self.father else None
        if family:
            data["family"] = self.case.to_dict()
        return data

    def __str__(self) -> str:
        return f"{self.case.internal_id} | {self.sample.internal_id}"

    @property
    def get_maternal_sample_id(self) -> str:
        """Return parental id."""
        return self.mother.internal_id if self.mother else EMPTY_STRING

    @property
    def get_paternal_sample_id(self) -> str:
        """Return parental id."""
        return self.father.internal_id if self.father else EMPTY_STRING


class Flowcell(Base):
    __tablename__ = "flowcell"
    id: Mapped[PrimaryKeyInt]
    name: Mapped[UniqueStr]
    sequencer_type: Mapped[str | None] = mapped_column(
        types.Enum("hiseqga", "hiseqx", "novaseq", "novaseqx")
    )
    sequencer_name: Mapped[Str32 | None]
    sequenced_at: Mapped[datetime | None]
    status: Mapped[str | None] = mapped_column(
        types.Enum(*(status.value for status in SequencingRunDataAvailability)), default="ondisk"
    )
    archived_at: Mapped[datetime | None]
    has_backup: Mapped[bool] = mapped_column(default=False)
    updated_at: Mapped[datetime | None] = mapped_column(onupdate=datetime.now)
    sequencing_metrics: Mapped[list["SampleLaneSequencingMetrics"]] = orm.relationship(
        back_populates="flowcell",
        cascade="all, delete, delete-orphan",
    )

    @property
    def samples(self) -> list["Sample"]:
        """Return samples sequenced on the flow cell."""
        return list({metric.sample for metric in self.sequencing_metrics})

    def __str__(self):
        return self.name

    def to_dict(self, samples: bool = False):
        """Represent as dictionary"""
        data = to_dict(model_instance=Flowcell)
        if samples:
            data["samples"] = [sample.to_dict() for sample in self.samples]
        return data


class Organism(Base):
    __tablename__ = "organism"
    id: Mapped[PrimaryKeyInt]
    internal_id: Mapped[UniqueStr]
    name: Mapped[Str255] = mapped_column(unique=True)
    created_at: Mapped[datetime | None] = mapped_column(default=datetime.now)
    updated_at: Mapped[datetime | None] = mapped_column(onupdate=datetime.now)
    reference_genome: Mapped[Str255 | None]
    verified: Mapped[bool | None] = mapped_column(default=False)
    comment: Mapped[Text | None]

    def __str__(self) -> str:
        return f"{self.internal_id} ({self.name})"

    def to_dict(self) -> dict:
        """Represent as dictionary"""
        return to_dict(model_instance=self)


class Panel(Base):
    __tablename__ = "panel"
    abbrev: Mapped[UniqueStr | None]
    current_version: Mapped[float]
    customer_id: Mapped[int] = mapped_column(ForeignKey("customer.id", ondelete="CASCADE"))
    customer: Mapped[Customer] = orm.relationship(back_populates="panels")
    date: Mapped[datetime]
    gene_count: Mapped[int | None]
    id: Mapped[PrimaryKeyInt]
    name: Mapped[Str64 | None] = mapped_column(unique=True)

    def __str__(self):
        return f"{self.abbrev} ({self.current_version})"

    def to_dict(self):
        return to_dict(model_instance=self)


class Pool(Base):
    __tablename__ = "pool"
    __table_args__ = (UniqueConstraint("order", "name", name="_order_name_uc"),)

    application_version_id: Mapped[int] = mapped_column(ForeignKey("application_version.id"))
    application_version: Mapped["ApplicationVersion"] = orm.relationship(
        foreign_keys=[application_version_id]
    )
    comment: Mapped[Text | None]
    created_at: Mapped[datetime | None] = mapped_column(default=datetime.now)
    customer_id: Mapped[int] = mapped_column(ForeignKey("customer.id", ondelete="CASCADE"))
    customer: Mapped[Customer] = orm.relationship(foreign_keys=[customer_id])
    delivered_at: Mapped[datetime | None]
    id: Mapped[PrimaryKeyInt]
    invoice_id: Mapped[int | None] = mapped_column(ForeignKey("invoice.id"))
    name: Mapped[Str32]
    no_invoice: Mapped[bool | None] = mapped_column(default=False)
    order: Mapped[Str64]
    ordered_at: Mapped[datetime]
    received_at: Mapped[datetime | None]
    ticket: Mapped[Str32 | None]

    invoice: Mapped["Invoice | None"] = orm.relationship(back_populates="pools")

    def to_dict(self):
        return to_dict(model_instance=self)


class Sample(Base, PriorityMixin):
    __tablename__ = "sample"
    age_at_sampling: Mapped[float | None]
    application_version_id: Mapped[int] = mapped_column(ForeignKey("application_version.id"))
    application_version: Mapped[ApplicationVersion] = orm.relationship(
        foreign_keys=[application_version_id]
    )
    is_cancelled: Mapped[bool] = mapped_column(default=False, nullable=False)
    capture_kit: Mapped[Str64 | None]
    comment: Mapped[Text | None]
    control: Mapped[str | None] = mapped_column(
        types.Enum(*(option.value for option in ControlOptions))
    )
    created_at: Mapped[datetime | None] = mapped_column(default=datetime.now)
    customer_id: Mapped[int] = mapped_column(ForeignKey("customer.id", ondelete="CASCADE"))
    customer: Mapped[Customer] = orm.relationship(foreign_keys=[customer_id])
    delivered_at: Mapped[datetime | None]
    downsampled_to: Mapped[BigInt | None]
    from_sample: Mapped[Str128 | None]
    id: Mapped[PrimaryKeyInt]
    internal_id: Mapped[UniqueStr]
    invoice_id: Mapped[int | None] = mapped_column(ForeignKey("invoice.id"))
    is_tumour: Mapped[bool | None] = mapped_column(default=False)
    loqusdb_id: Mapped[Str64 | None]
    name: Mapped[Str128]
    no_invoice: Mapped[bool] = mapped_column(default=False)
    order: Mapped[Str64 | None]
    ordered_at: Mapped[datetime]
    organism_id: Mapped[int | None] = mapped_column(ForeignKey("organism.id"))
    organism: Mapped["Organism | None"] = orm.relationship(foreign_keys=[organism_id])
    original_ticket: Mapped[Str32 | None]
    _phenotype_groups: Mapped[str | None] = mapped_column(types.Text)
    _phenotype_terms: Mapped[str | None] = mapped_column(types.Text)
    prepared_at: Mapped[datetime | None]

    priority: Mapped[Priority] = mapped_column(default=Priority.standard)
    reads: Mapped[BigInt | None] = mapped_column(default=0)
    last_sequenced_at: Mapped[datetime | None]
    received_at: Mapped[datetime | None]
    reference_genome: Mapped[Str255 | None]
    sex: Mapped[str] = mapped_column(types.Enum(*(option.value for option in SexOptions)))
    subject_id: Mapped[Str128 | None]

    links: Mapped[list[CaseSample]] = orm.relationship(
        foreign_keys=[CaseSample.sample_id], back_populates="sample"
    )
    mother_links: Mapped[list[CaseSample]] = orm.relationship(
        foreign_keys=[CaseSample.mother_id], back_populates="mother"
    )
    father_links: Mapped[list[CaseSample]] = orm.relationship(
        foreign_keys=[CaseSample.father_id], back_populates="father"
    )
    sequencing_metrics: Mapped[list["SampleLaneSequencingMetrics"]] = orm.relationship(
        back_populates="sample"
    )
    invoice: Mapped["Invoice | None"] = orm.relationship(back_populates="samples")

    _sample_run_metrics: Mapped[list["SampleRunMetrics"]] = orm.relationship(
        back_populates="sample", cascade="all, delete"
    )

    def __str__(self) -> str:
        return f"{self.internal_id} ({self.name})"

    @property
    def archive_location(self) -> str:
        """Returns the data_archive_location if the customer linked to the sample."""
        return self.customer.data_archive_location

    @property
    def expected_reads_for_sample(self) -> int:
        """Return the expected reads of the sample."""
        return self.application_version.application.expected_reads

    @property
    def has_reads(self) -> bool:
        return bool(self.reads)

    @property
    def flow_cells(self) -> list[Flowcell]:
        """Return the flow cells a sample has been sequenced on."""
        return list({metric.flowcell for metric in self.sequencing_metrics})

    @property
    def phenotype_groups(self) -> list[str]:
        """Return a list of phenotype_groups."""
        return self._phenotype_groups.split(",") if self._phenotype_groups else []

    @phenotype_groups.setter
    def phenotype_groups(self, phenotype_term_list: list[str]):
        self._phenotype_groups = ",".join(phenotype_term_list) if phenotype_term_list else None

    @property
    def phenotype_terms(self) -> list[str]:
        """Return a list of phenotype_terms."""
        return self._phenotype_terms.split(",") if self._phenotype_terms else []

    @phenotype_terms.setter
    def phenotype_terms(self, phenotype_term_list: list[str]):
        self._phenotype_terms = ",".join(phenotype_term_list) if phenotype_term_list else None

    @property
    def prep_category(self) -> str:
        """Return the preparation category of the sample."""
        return self.application_version.application.prep_category

    @property
    def state(self) -> str:
        """Get the current sample state."""
        if self.delivered_at:
            return f"Delivered {self.delivered_at.date()}"
        if self.last_sequenced_at:
            return f"Sequenced {self.last_sequenced_at.date()}"
        if self.received_at:
            return f"Received {self.received_at.date()}"

        return f"Ordered {self.ordered_at.date()}"

    @property
    def run_devices(self) -> list["RunDevice"]:
        """Return the run_devices a sample has been sequenced on."""
        return list({metric.instrument_run.device for metric in self._sample_run_metrics})

    @property
    def sample_run_metrics(self) -> list["SampleRunMetrics"]:
        """Return the sample run metrics for the sample."""
        return self._sample_run_metrics

    def to_dict(self, links: bool = False, flowcells: bool = False) -> dict:
        """Represent as dictionary"""
        data = to_dict(model_instance=self)
        data["priority"] = self.priority_human
        data["customer"] = self.customer.to_dict()
        data["application_version"] = self.application_version.to_dict()
        data["application"] = self.application_version.application.to_dict()
        if links:
            data["links"] = [link_obj.to_dict(family=True, parents=True) for link_obj in self.links]
        if flowcells:
            data["flowcells"] = [flow_cell.to_dict() for flow_cell in self.flow_cells]
        return data


class Invoice(Base):
    __tablename__ = "invoice"
    id: Mapped[PrimaryKeyInt]
    customer_id: Mapped[int] = mapped_column(ForeignKey("customer.id"))
    customer: Mapped[Customer] = orm.relationship(foreign_keys=[customer_id])
    created_at: Mapped[datetime | None] = mapped_column(default=datetime.now)
    updated_at: Mapped[datetime | None] = mapped_column(onupdate=datetime.now)
    invoiced_at: Mapped[datetime | None]
    comment: Mapped[Text | None]
    discount: Mapped[int | None] = mapped_column(default=0)
    excel_kth: Mapped[Blob | None]
    excel_ki: Mapped[Blob | None]
    price: Mapped[int | None]
    record_type: Mapped[Text | None]

    samples: Mapped[list["Sample"]] = orm.relationship(back_populates="invoice")
    pools: Mapped[list["Pool"]] = orm.relationship(back_populates="invoice")

    def __str__(self):
        return f"{self.customer_id} ({self.invoiced_at})"

    def to_dict(self) -> dict:
        """Represent as dictionary"""
        return to_dict(model_instance=self)


class User(Base):
    __tablename__ = "user"
    id: Mapped[PrimaryKeyInt]
    name: Mapped[Str128]
    email: Mapped[Str128] = mapped_column(unique=True)
    is_admin: Mapped[bool | None] = mapped_column(default=False)
    order_portal_login: Mapped[bool | None] = mapped_column(default=False)

    customers: Mapped[list[Customer]] = orm.relationship(
        secondary=customer_user, back_populates="users"
    )

    def to_dict(self) -> dict:
        """Represent as dictionary."""
        dict_representation: dict = to_dict(model_instance=self)
        dict_representation["customers"] = [customer.to_dict() for customer in self.customers]
        return dict_representation

    def __str__(self) -> str:
        return self.name


class SampleLaneSequencingMetrics(Base):
    """Model for storing sequencing metrics per lane and sample."""

    __tablename__ = "sample_lane_sequencing_metrics"

    id: Mapped[PrimaryKeyInt]
    flow_cell_name: Mapped[Str32] = mapped_column(ForeignKey("flowcell.name"))
    flow_cell_lane_number: Mapped[int | None]

    sample_internal_id: Mapped[Str32] = mapped_column(ForeignKey("sample.internal_id"))
    sample_total_reads_in_lane: Mapped[BigInt | None]
    sample_base_percentage_passing_q30: Mapped[Num_6_2 | None]
    sample_base_mean_quality_score: Mapped[Num_6_2 | None]

    created_at: Mapped[datetime | None]

    flowcell: Mapped[Flowcell] = orm.relationship(back_populates="sequencing_metrics")
    sample: Mapped[Sample] = orm.relationship(back_populates="sequencing_metrics")

    __table_args__ = (
        UniqueConstraint(
            "flow_cell_name",
            "sample_internal_id",
            "flow_cell_lane_number",
            name="uix_flowcell_sample_lane",
        ),
    )

    def to_dict(self):
        return to_dict(model_instance=self)


class Order(Base):
    """Model for storing orders."""

    __tablename__ = "order"

    id: Mapped[PrimaryKeyInt]
    cases: Mapped[list[Case]] = orm.relationship(secondary=order_case, back_populates="orders")
    customer_id: Mapped[int] = mapped_column(ForeignKey("customer.id"))
    customer: Mapped[Customer] = orm.relationship(foreign_keys=[customer_id])
    order_date: Mapped[datetime] = mapped_column(default=datetime.now())
    ticket_id: Mapped[int] = mapped_column(unique=True, index=True)
    workflow: Mapped[str] = mapped_column(types.Enum(*(workflow.value for workflow in Workflow)))
    is_delivered: Mapped[bool] = mapped_column(default=False)

    def to_dict(self):
        return to_dict(model_instance=self)


class RunDevice(Base):
    """Parent model for the different types of run run_devices."""

    __tablename__ = "run_device"

    id: Mapped[PrimaryKeyInt]
    type: Mapped[DeviceType]
    internal_id: Mapped[UniqueStr64]

    instrument_runs: Mapped[list["InstrumentRun"]] = orm.relationship(
        back_populates="device", cascade="all, delete"
    )

    @property
    def samples(self) -> list[Sample]:
        """Return the samples sequenced in this device."""
        return list(
            {
                sample_run_metric.sample
                for run in self.instrument_runs
                for sample_run_metric in run.sample_metrics
            }
        )

    __mapper_args__ = {
        "polymorphic_on": "type",
    }


class IlluminaFlowCell(RunDevice):
    """Model for storing Illumina flow cells."""

    __tablename__ = "illumina_flow_cell"

    id: Mapped[int] = mapped_column(ForeignKey("run_device.id"), primary_key=True)
    model: Mapped[str | None] = mapped_column(
        types.Enum("10B", "25B", "1.5B", "S1", "S2", "S4", "SP")
    )

    __mapper_args__ = {"polymorphic_identity": DeviceType.ILLUMINA}


class PacBioSMRTCell(RunDevice):
    """Model for storing PacBio SMRT cells."""

    __tablename__ = "pacbio_smrt_cell"

    id: Mapped[int] = mapped_column(ForeignKey("run_device.id"), primary_key=True)

    __mapper_args__ = {"polymorphic_identity": DeviceType.PACBIO}


class InstrumentRun(Base):
    """Parent model for the different types of instrument runs."""

    __tablename__ = "instrument_run"

    id: Mapped[PrimaryKeyInt]
    type: Mapped[DeviceType]
    device_id: Mapped[int] = mapped_column(ForeignKey("run_device.id"))

    device: Mapped[RunDevice] = orm.relationship(back_populates="instrument_runs")
    sample_metrics: Mapped[list["SampleRunMetrics"]] = orm.relationship(
        back_populates="instrument_run", cascade="all, delete"
    )

    __mapper_args__ = {
        "polymorphic_on": "type",
    }


class IlluminaSequencingRun(InstrumentRun):
    __tablename__ = "illumina_sequencing_run"

    id: Mapped[int] = mapped_column(ForeignKey("instrument_run.id"), primary_key=True)
    sequencer_type: Mapped[str | None] = mapped_column(
        types.Enum("hiseqga", "hiseqx", "novaseq", "novaseqx")
    )
    sequencer_name: Mapped[Str32 | None]
    data_availability: Mapped[str | None] = mapped_column(
        types.Enum(*(status.value for status in SequencingRunDataAvailability)), default="ondisk"
    )
    archived_at: Mapped[datetime | None]
    has_backup: Mapped[bool] = mapped_column(default=False)
    total_reads: Mapped[BigInt | None]
    total_undetermined_reads: Mapped[BigInt | None]
    percent_undetermined_reads: Mapped[Num_6_2 | None]
    percent_q30: Mapped[Num_6_2 | None]
    mean_quality_score: Mapped[Num_6_2 | None]
    total_yield: Mapped[BigInt | None]
    yield_q30: Mapped[BigInt | None]
    cycles: Mapped[int | None]
    demultiplexing_software: Mapped[Str32 | None]
    demultiplexing_software_version: Mapped[Str32 | None]
    sequencing_started_at: Mapped[datetime | None]
    sequencing_completed_at: Mapped[datetime | None]
    demultiplexing_started_at: Mapped[datetime | None]
    demultiplexing_completed_at: Mapped[datetime | None]

    __mapper_args__ = {"polymorphic_identity": DeviceType.ILLUMINA}

    def to_dict(self):
        """Represent as dictionary"""
        data = to_dict(model_instance=IlluminaSequencingRun)
        return data


class PacBioSequencingRun(InstrumentRun):
    __tablename__ = "pacbio_sequencing_run"

    id: Mapped[int] = mapped_column(ForeignKey("instrument_run.id"), primary_key=True)
    well: Mapped[Str32]
    plate: Mapped[int]
    movie_time_hours: Mapped[int]
    movie_name: Mapped[Str32]
    hifi_reads: Mapped[BigInt]
    hifi_yield: Mapped[BigInt]
    hifi_mean_read_length: Mapped[BigInt]
    hifi_median_read_quality: Mapped[Str32]
    percent_reads_passing_q30: Mapped[Num_6_2]
    p0_percent: Mapped[Num_6_2]
    p1_percent: Mapped[Num_6_2]
    p2_percent: Mapped[Num_6_2]
    productive_zmws: Mapped[BigInt]
    polymerase_mean_read_length: Mapped[BigInt]
    polymerase_read_length_n50: Mapped[BigInt]
    polymerase_mean_longest_subread: Mapped[BigInt]
    polymerase_longest_subread_n50: Mapped[BigInt]
    control_reads: Mapped[BigInt]
    control_mean_read_length: Mapped[BigInt]
    control_mean_read_concordance: Mapped[Num_6_2]
    control_mode_read_concordance: Mapped[Num_6_2]
    failed_reads: Mapped[BigInt]
    failed_yield: Mapped[BigInt]
    failed_mean_read_length: Mapped[BigInt]

    __mapper_args__ = {"polymorphic_identity": DeviceType.PACBIO}


class SampleRunMetrics(Base):
    """Parent model for the different types of sample run metrics."""

    __tablename__ = "sample_run_metrics"
    id: Mapped[PrimaryKeyInt]
    sample_id: Mapped[int] = mapped_column(ForeignKey("sample.id"))
    instrument_run_id: Mapped[int] = mapped_column(ForeignKey("instrument_run.id"))
    type: Mapped[DeviceType]

    instrument_run: Mapped[InstrumentRun] = orm.relationship(back_populates="sample_metrics")
    sample: Mapped[Sample] = orm.relationship(back_populates="_sample_run_metrics")

    __mapper_args__ = {
        "polymorphic_on": "type",
    }


class IlluminaSampleSequencingMetrics(SampleRunMetrics):
    """Sequencing metrics for a sample sequenced on an Illumina instrument. The metrics are per sample, per lane, per flow cell."""

    __tablename__ = "illumina_sample_sequencing_metrics"

    id: Mapped[int] = mapped_column(ForeignKey("sample_run_metrics.id"), primary_key=True)
    flow_cell_lane: Mapped[int | None]
    total_reads_in_lane: Mapped[BigInt | None]
    base_passing_q30_percent: Mapped[Num_6_2 | None]
    base_mean_quality_score: Mapped[Num_6_2 | None]
    yield_: Mapped[BigInt | None] = mapped_column("yield", quote=True)
    yield_q30: Mapped[Num_6_2 | None]
    created_at: Mapped[datetime | None]
    __mapper_args__ = {"polymorphic_identity": DeviceType.ILLUMINA}


class PacBioSampleSequencingMetrics(SampleRunMetrics):
    """Sequencing metrics for a sample sequenced on a PacBio instrument. The metrics are per sample, per cell."""

    __tablename__ = "pacbio_sample_sequencing_metrics"

    id: Mapped[int] = mapped_column(ForeignKey("sample_run_metrics.id"), primary_key=True)
    hifi_reads: Mapped[BigInt]
    hifi_yield: Mapped[BigInt]
    hifi_mean_read_length: Mapped[BigInt]
    hifi_median_read_quality: Mapped[Str32]
    percent_reads_passing_q30: Mapped[Num_6_2]
    failed_reads: Mapped[BigInt]
    failed_yield: Mapped[BigInt]
    failed_mean_read_length: Mapped[BigInt]

    __mapper_args__ = {"polymorphic_identity": DeviceType.PACBIO}



from pathlib import Path

import pkg_resources

from cg.constants import FileExtensions

RAREDISEASE_BUNDLE_FILENAMES: str = (
    Path("resources", "raredisease_bundle_filenames").with_suffix(FileExtensions.YAML).as_posix()
)

RNAFUSION_BUNDLE_FILENAMES: str = (
    Path("resources", "rnafusion_bundle_filenames").with_suffix(FileExtensions.YAML).as_posix()
)

TAXPROFILER_BUNDLE_FILENAMES: str = (
    Path("resources", "taxprofiler_bundle_filenames").with_suffix(FileExtensions.YAML).as_posix()
)

TOMTE_BUNDLE_FILENAMES: str = (
    Path("resources", "tomte_bundle_filenames").with_suffix(FileExtensions.YAML).as_posix()
)

RAREDISEASE_BUNDLE_FILENAMES_PATH = Path(
    pkg_resources.resource_filename("cg", RAREDISEASE_BUNDLE_FILENAMES)
)

RNAFUSION_BUNDLE_FILENAMES_PATH: Path = Path(
    pkg_resources.resource_filename("cg", RNAFUSION_BUNDLE_FILENAMES)
)

TAXPROFILER_BUNDLE_FILENAMES_PATH: Path = Path(
    pkg_resources.resource_filename("cg", TAXPROFILER_BUNDLE_FILENAMES)
)

TOMTE_BUNDLE_FILENAMES_PATH: Path = Path(
    pkg_resources.resource_filename("cg", TOMTE_BUNDLE_FILENAMES)
)



"""Module for the NovaSeqX sequencing times service."""

from datetime import datetime
from pathlib import Path

from cg.models.run_devices.illumina_run_directory_data import IlluminaRunDirectoryData

from cg.constants.demultiplexing import RunCompletionStatusNodes
from cg.io.xml import read_xml
from cg.services.illumina.file_parsing.sequencing_times.sequencing_time_service import (
    SequencingTimesService,
)
from cg.utils.time import format_time_from_string
from xml.etree.ElementTree import ElementTree


class NovaseqXSequencingTimesService(SequencingTimesService):
    """Class to get the sequencing times for NovaSeqX sequencing runs."""

    @staticmethod
    def get_start_time(run_directory_data: IlluminaRunDirectoryData) -> datetime:
        """Get the sequencer start date and time."""
        run_completion_status_path: Path = run_directory_data.get_run_completion_status()
        tree: ElementTree = read_xml(run_completion_status_path)
        return format_time_from_string(tree.find(RunCompletionStatusNodes.RUN_START).text)

    @staticmethod
    def get_end_time(run_directory_data: IlluminaRunDirectoryData) -> datetime:
        """Get the sequencer end date and time."""
        run_completion_status_path: Path = run_directory_data.get_run_completion_status()
        tree: ElementTree = read_xml(run_completion_status_path)
        return format_time_from_string(tree.find(RunCompletionStatusNodes.RUN_END).text)
